# -*- coding: utf-8 -*-
import logging
from typing import Any, List

_logger = logging.getLogger(__name__)  # type: ignore


def flatten_list(lst: List[Any]) -> List[Any]:
    items = list()
    for item in lst:
        if isinstance(item, list):
            if len(item) > 2 and item[0] == "[" and item[-1] == "]":
                items.append(
                    "[%s]" % (",".join(str(item[i]) for i in range(1, len(item) - 1)))
                )
                return items

            items.extend(flatten_list(item))
        else:
            items.append(item)
    return items



# -*- coding: utf-8 -*-
import logging
from .base import Base
from .common import KeyWords, Tokens
from pyparsing import (
    CaselessKeyword,
    Opt,
    Group,
    ZeroOrMore,
    OneOrMore,
    one_of,
    delimited_list,
    ParseResults,
    ParserElement,
)
from typing import Any, Dict, List, Optional

_logger = logging.getLogger(__name__)  # type: ignore


class DdlBase(Base):
    ParserElement.enablePackrat()

    _PROJECTION_PROJECTIONTYPE, _PROJECTION_NONKEYATTRIBUTES = map(
        CaselessKeyword,
        [
            "Projection.ProjectionType",
            "Projection.NonKeyAttributes",
        ],
    )

    (
        _PROVISIONEDTHROUGHPUT_READCAPACITYUNITS,
        _PROVISIONEDTHROUGHPUT_WRITECAPACITYUNITS,
        _BILLINGMODE,
        _STREAMSPECIFICATION_STREAMENABLED,
        _STREAMSPECIFICATION_STREAMVIEWTYPE,
        _SSESPECIFICATION_ENABLED,
        _SSESPECIFICATION_SSETYPE,
        _SSESPECIFICATION_KMSMASTERKEYID,
        _KMSMASTERKEYID,
        _TABLECLASS,
        _TAGS,
        _REPLICATION_GROUP,
    ) = map(
        CaselessKeyword,
        [
            "ProvisionedThroughput.ReadCapacityUnits",
            "ProvisionedThroughput.WriteCapacityUnits",
            "BillingMode",
            "StreamSpecification.StreamEnabled",
            "StreamSpecification.StreamViewType",
            "SSESpecification.Enabled",
            "SSESpecification.SSEType",
            "SSESpecification.KMSMasterKeyId",
            "KMSMasterKeyId",
            "TableClass",
            "Tags",
            "ReplicationGroup",
        ],
    )

    (
        _GLOBALSECONDARYINDEXES,
        _PROVISIONEDTHROUGHPUTOVERRIDE_READCAPACITYUNITS,
        _TABLECLASSOVERRIDE,
    ) = map(
        CaselessKeyword,
        [
            "GlobalSecondaryIndexes",
            "ProvisionedThroughputOverride.ReadCapacityUnits",
            "TableClassOverride",
        ],
    )

    _TABLE_OPTIONS = ZeroOrMore(
        Group(
            _PROVISIONEDTHROUGHPUT_READCAPACITYUNITS
            + Opt(KeyWords.EQUALS)
            + Tokens.INT_VALUE
            ^ _PROVISIONEDTHROUGHPUT_WRITECAPACITYUNITS
            + Opt(KeyWords.EQUALS)
            + Tokens.INT_VALUE
            ^ _BILLINGMODE
            + Opt(KeyWords.EQUALS)
            + one_of("PROVISIONED PAY_PER_REQUEST")
            ^ _STREAMSPECIFICATION_STREAMENABLED
            + Opt(KeyWords.EQUALS)
            + Tokens.BOOL_VALUE
            ^ _STREAMSPECIFICATION_STREAMVIEWTYPE
            + Opt(KeyWords.EQUALS)
            + one_of("NEW_IMAGE OLD_IMAGE NEW_AND_OLD_IMAGES KEYS_ONLY")
            ^ _SSESPECIFICATION_ENABLED + Opt(KeyWords.EQUALS) + Tokens.BOOL_VALUE
            ^ _SSESPECIFICATION_SSETYPE + Opt(KeyWords.EQUALS) + one_of("AES256 KMS")
            ^ _SSESPECIFICATION_KMSMASTERKEYID
            + Opt(KeyWords.EQUALS)
            + Tokens.KMSMASTERKEYID
            ^ _TABLECLASS
            + Opt(KeyWords.EQUALS)
            + one_of("STANDARD STANDARD_INFREQUENT_ACCESS")
            ^ _TAGS
            + Opt(KeyWords.EQUALS)
            + KeyWords.LPAR
            + Group(delimited_list(Tokens.TAG))("tags").set_name("tags")
            + KeyWords.RPAR
            ^ _REPLICATION_GROUP
            + KeyWords.LPAR
            + Group(delimited_list(Tokens.REGION_NAME))("replication_group").set_name(
                "replication_group"
            )
            + KeyWords.RPAR
        )("table_option").set_name("table_option")
    )("table_options").set_name("table_options")

    _INDEX_OPTIONS = ZeroOrMore(
        Group(
            _PROVISIONEDTHROUGHPUT_READCAPACITYUNITS
            + Opt(KeyWords.EQUALS)
            + Tokens.INT_VALUE
            ^ _PROVISIONEDTHROUGHPUT_WRITECAPACITYUNITS
            + Opt(KeyWords.EQUALS)
            + Tokens.INT_VALUE
            ^ _PROJECTION_PROJECTIONTYPE
            + Opt(KeyWords.EQUALS)
            + one_of("ALL KEYS_ONLY INCLUDE")
            ^ _PROJECTION_NONKEYATTRIBUTES
            + Opt(KeyWords.EQUALS)
            + KeyWords.LPAR
            + Group(delimited_list(Tokens.ATTRIBUTE_NAME))(
                "nonkey_attributes"
            ).set_name("nonkey_attributes")
            + KeyWords.RPAR
        )("index_option").set_name("index_option")
    )("index_options").set_name("index_options")

    _INDEX = Group(
        KeyWords.INDEX
        + Tokens.INDEX_NAME
        + Tokens.INDEX_TYPE
        + KeyWords.LPAR
        + delimited_list(OneOrMore(Group(Tokens.ATTRIBUTE_NAME + Tokens.KEY_TYPE)))(
            "attributes"
        ).set_name("attributes")
        + KeyWords.RPAR
        + _INDEX_OPTIONS
    )("index").set_name("index")

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    def _construct_attr_def(self, attr_name: str, data_type: str) -> Dict[str, str]:
        converted_data_type = None

        if data_type == "NUMERIC":
            converted_data_type = "N"
        elif data_type == "STRING":
            converted_data_type = "S"
        elif data_type == "BINARY":
            converted_data_type = "B"
        else:
            raise LookupError("Data type is invalid")

        return {"AttributeName": attr_name, "AttributeType": converted_data_type}

    def _construct_key_schema(self, attr_name: str, key_type: str) -> Dict[str, str]:
        converted_key_type = None
        if key_type == "PARTITION KEY" or key_type == "HASH":
            converted_key_type = "HASH"
        elif key_type == "SORT KEY" or key_type == "RANGE":
            converted_key_type = "RANGE"
        else:
            raise LookupError("Key schema type is invalid")

        return {"AttributeName": attr_name, "KeyType": converted_key_type}

    def _construct_index(self, index: ParseResults) -> Dict[str, Any]:
        index_ = dict()
        index_name = index["index_name"]
        index_.update({"IndexName": index_name})

        attrs = index.get("attributes", None)
        if attrs:
            key_schema = list()

            for attr in attrs:
                attr_name = attr["attribute_name"]
                key_type = attr["key_type"]

                key_schema.append(self._construct_key_schema(attr_name, key_type))
            index_.update({"KeySchema": key_schema})

        options = index["index_options"]
        index_.update(self._construct_options(options))
        return index_

    def _construct_options(self, options: List[Any]) -> Optional[Dict[str, Any]]:
        converted_ = None
        for option in options:
            if converted_ is None:
                converted_ = dict()
            option_name_path = option[0]
            option_value = option[1]

            if isinstance(option_value, ParseResults):
                if (
                    option.get_name() == "index_option"
                    and option_value.get_name() == "nonkey_attributes"
                ):
                    option_value = [o for o in option_value]
                elif (
                    option.get_name() == "table_option"
                    and option_value.get_name() == "tags"
                ):
                    option_value = [
                        {"Key": o["tag_key"], "Value": o["tag_value"]}
                        for o in option_value
                    ]
                elif (
                    option.get_name() == "table_option"
                    and option_value.get_name() == "replication_group"
                ):
                    option_value = [{"RegionName": o} for o in option_value]

            self._parse_option_path(option_name_path, option_value, converted_)
        return converted_

    def _parse_option_path(
        self,
        option_name_path: str,
        option_value: Any,
        return_dict: Dict[str, Any] = None,
    ) -> Dict[str, Any]:
        if return_dict is None:
            return_dict = dict()

        names = option_name_path.split(".")
        names_size = len(names)

        if names_size == 1:
            return_dict.update({names[0]: option_value})
        elif names_size == 2:
            if names[0] in return_dict:
                return_dict[names[0]].update({names[1]: option_value})
            else:
                return_dict.update({names[0]: {names[1]: option_value}})
        else:
            """Ignore the option"""
            pass

        return return_dict



# -*- coding: utf-8 -*-
import hashlib
import logging
import re
from datetime import date, datetime, timedelta
from contextlib import closing
from abc import ABCMeta, abstractmethod
from typing import Tuple, Optional, Type, Any, List

from ..sql.common import DataTypes, QueryType, Functions
from ..sql.dml_select import DmlSelectColumn, DmlFunction
from ..model import Statement, Metadata
from ..util import synchronized
from ..error import NotSupportedError

_logger = logging.getLogger(__name__)  # type: ignore


DEFAULT_QUERYDB_TYPE = "sqlite"
DEFAULT_QUERYDB_URL = ":memory:"
DEFAULT_QUERYDB_LOAD_BATCH_SIZE = 200
DEFAULT_QUERYDB_EXPIRE_TIME = 300
DEFAULT_QUERYDB_PURGE_TIME = 86400


SUPPORTED_QUERYDB_CONFIG = {
    "querydb_type": ("PYDYNAMODB_QUERYDB_TYPE", DEFAULT_QUERYDB_TYPE),
    "querydb_class": ("PYDYNAMODB_QUERYDB_CLASS", None),
    "querydb_url": ("PYDYNAMODB_QUERYDB_URL", DEFAULT_QUERYDB_URL),
    "querydb_load_batch_size": (
        "PYDYNAMODB_QUERYDB_LOAD_BATCH_SIZE",
        DEFAULT_QUERYDB_LOAD_BATCH_SIZE,
    ),
    "querydb_expire_time": (
        "PYDYNAMODB_QUERYDB_EXPIRE_TIME",
        DEFAULT_QUERYDB_EXPIRE_TIME,
    ),
    "querydb_purge_enabled": (
        "PYDYNAMODB_QUERYDB_PURGE_ENABLED",
        "True",
    ),
    "querydb_purge_time": (
        "PYDYNAMODB_QUERYDB_PURGE_TIME",
        DEFAULT_QUERYDB_PURGE_TIME,
    ),
}


class QueryDBConfig:
    def __init__(
        self,
        db_type: str,
        db_class: str,
        db_url: str,
        load_batch_size: int = DEFAULT_QUERYDB_LOAD_BATCH_SIZE,
        expire_time: int = DEFAULT_QUERYDB_EXPIRE_TIME,
        purge_enabled: bool = True,
        purge_time: int = DEFAULT_QUERYDB_PURGE_TIME,
    ):
        self._db_type = db_type
        self._db_url = db_url
        self._db_class = db_class
        self._load_batch_size = load_batch_size
        self._expire_time = expire_time
        self._purge_enabled = purge_enabled
        self._purge_time = purge_time

    @property
    def db_type(self) -> str:
        return self._db_type

    @property
    def db_class(self) -> str:
        return self._db_class

    @property
    def db_url(self) -> str:
        return self._db_url

    @property
    def load_batch_size(self) -> int:
        return self._load_batch_size

    @property
    def expire_time(self) -> int:
        return self._expire_time

    @property
    def purge_enabled(self) -> int:
        return self._purge_enabled

    @property
    def purge_time(self) -> int:
        return self._purge_time


class QueryDB(metaclass=ABCMeta):
    CACHE_TABLE = "QUERYDB_CACHES"

    def __init__(
        self,
        statement: Statement,
        config: QueryDBConfig,
        **kwargs,
    ) -> None:
        self._config = config
        self._statement = statement
        cache_enabled_ = kwargs.get("cache_enabled", None)
        self._cache_enabled = cache_enabled_ if cache_enabled_ is not None else True
        self._kwargs = kwargs
        self._query_id = None

    @property
    def statement(self) -> Statement:
        return self._statement

    @property
    def config(self) -> QueryDBConfig:
        return self._config

    @property
    def cache_enabled(self) -> bool:
        return self._cache_enabled

    @property
    def query_id(self) -> str:
        if self._query_id is None:
            sql_parser = self._statement.sql_parser
            if sql_parser.query_type == QueryType.SELECT:
                columns = ",".join([str(c) for c in sql_parser.parser.columns])
                hash_key = str(self._statement.api_request) + columns
                self._query_id = (
                    "_T" + hashlib.md5(hash_key.encode("utf-8")).hexdigest()
                )
            else:
                raise NotSupportedError

        return self._query_id

    @abstractmethod
    def connection(self):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def type_conversion(self, type: Type[Any]) -> str:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def has_table(self, table: str) -> bool:
        raise NotImplementedError  # pragma: no cover

    def rollback(self):
        if self.connection is not None:
            self.connection.rollback()

    def close(self):
        if self.connection is not None:
            if self.cache_enabled and self.config.purge_enabled:
                self.purge()
            self.connection.close()

    @synchronized
    def purge(self) -> int:
        purged_count = 0

        if self.config.purge_time < self.config.expire_time:
            return purged_count

        if not self.cache_enabled or not self.has_table(QueryDB.CACHE_TABLE):
            return purged_count

        _purge_period = datetime.now() - timedelta(seconds=self.config.purge_time)

        with closing(self.connection.cursor()) as cursor:
            cursor.execute(
                "SELECT query_id FROM %s WHERE last_updated<=?" % QueryDB.CACHE_TABLE,
                (_purge_period,),
            )
            result = cursor.fetchall()
            purged_count = len(result)
            for r in result:
                try:
                    cursor.execute("DROP TABLE %s" % (r[0]))
                except Exception:
                    _logger.warning("Failed to drop query table.")

            if purged_count > 0:
                cursor.executemany(
                    "DELETE FROM %s WHERE query_id=?" % QueryDB.CACHE_TABLE, result
                )
        self.connection.commit()

        return purged_count

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @synchronized
    def init_cache_table(self) -> None:
        if not self.cache_enabled:
            return

        if not self.has_table(QueryDB.CACHE_TABLE):
            columns = list()
            columns.append(
                "%s %s PRIMARY KEY" % ("query_id", self.type_conversion(str))
            )
            columns.append("%s %s" % ("statement", self.type_conversion(str)))
            columns.append("%s %s" % ("created", self.type_conversion(datetime)))
            columns.append("%s %s" % ("last_updated", self.type_conversion(datetime)))
            columns.append("%s %s" % ("queried_times", self.type_conversion(int)))
            self.connection.execute(
                "CREATE TABLE %s (%s)" % (QueryDB.CACHE_TABLE, ",".join(columns))
            )

    def get_cache(self) -> Optional[Tuple[Any]]:
        if not self.cache_enabled:
            return None

        with closing(self.connection.cursor()) as cursor:
            cursor.execute(
                "SELECT statement, created, last_updated FROM %s WHERE query_id=?"
                % QueryDB.CACHE_TABLE,
                (self.query_id,),
            )
            return cursor.fetchone()

    def has_cache(self) -> bool:
        if not self.cache_enabled:
            return False

        if not self.has_table(QueryDB.CACHE_TABLE):
            return False

        cache_ = self.get_cache()
        if cache_ is None:
            return False

        if (datetime.now() - cache_[2]).seconds <= self.config.expire_time:
            return True

        return False

    @synchronized
    def add_cache(self) -> None:
        if not self.cache_enabled:
            return

        with closing(self.connection.cursor()) as cursor:
            cursor.execute(
                "SELECT COUNT(query_id) FROM %s WHERE query_id=?" % QueryDB.CACHE_TABLE,
                (self.query_id,),
            )
            result = cursor.fetchone()
            if result is None or result == (0,):
                cursor.execute(
                    """INSERT INTO %s (query_id, statement,
                                    created, last_updated, queried_times
                        ) VALUES (?, ?, ?, ?, ?)
                    """
                    % QueryDB.CACHE_TABLE,
                    (
                        self.query_id,
                        str(self.statement.api_request),
                        datetime.now(),
                        datetime.now(),
                        0,
                    ),
                )
                self.connection.commit()

    @synchronized
    def set_cache_last_updated(self) -> None:
        if not self.cache_enabled:
            return

        with closing(self.connection.cursor()) as cursor:
            cursor.execute(
                "UPDATE %s SET last_updated=? WHERE query_id=?" % QueryDB.CACHE_TABLE,
                (datetime.now(), self.query_id),
            )
            self.connection.commit()

    @synchronized
    def set_cache_queried_times(self) -> None:
        if not self.cache_enabled:
            return

        with closing(self.connection.cursor()) as cursor:
            cursor.execute(
                "UPDATE %s SET queried_times=queried_times+1 WHERE query_id=?"
                % QueryDB.CACHE_TABLE,
                (self.query_id,),
            )
            self.connection.commit()

    def _get_col_type(self, col_info: DmlSelectColumn) -> str:
        col_type = self.type_conversion(str)

        if col_info.type_code == DataTypes.BOOL or col_info.type_code == DataTypes.NULL:
            col_type = self.type_conversion(int)
        elif col_info.type_code == DataTypes.NUMBER:
            col_type = self.type_conversion(float)
        elif col_info.type_code == DataTypes.DATE:
            col_type = self.type_conversion(date)
        elif col_info.type_code == DataTypes.DATETIME:
            col_type = self.type_conversion(datetime)
        return col_type

    def _get_col_function(self, col_name: str, col_function: DmlFunction) -> str:
        _col_func = col_name
        if (
            col_function is not None
            and col_function.name
            in Functions.SUPPORTED_FUNTIONS[Functions.STRING_FUNCTION]
        ):
            if col_function.params is not None:
                _col_func = '%s("%s", %s)' % (
                    col_function.name,
                    col_name,
                    ",".join(
                        f'"{x}"' if isinstance(x, str) else str(x)
                        for x in col_function.params
                    ),
                )
            else:
                _col_func = '%s("%s")' % (col_function.name, col_name)

        return _col_func

    def _escape_col_name(self, col_name: str) -> str:
        return re.sub(r"[^\w]", "_", col_name)

    def _get_query_table_col_name(self, col_info: DmlSelectColumn) -> str:
        return (
            self._escape_col_name(col_info.name)
            if col_info.alias is None
            else col_info.alias
        )

    @synchronized
    def create_query_table(self, metadata: Metadata) -> None:
        columns_with_type = list()

        for col_info in metadata:
            col_type = self._get_col_type(col_info)
            columns_with_type.append(
                '"%s" %s' % (self._get_query_table_col_name(col_info), col_type)
            )

        query_table_creation = "CREATE TABLE IF NOT EXISTS %s (%s)" % (
            self.query_id,
            ",".join(columns_with_type),
        )
        self.connection.execute(query_table_creation)
        self.add_cache()

    @synchronized
    def drop_query_table(self) -> None:
        if self.has_table(self.query_id):
            self.connection.execute("DROP TABLE %s" % (self.query_id))

    @synchronized
    def write_raw_data(
        self, metadata: Metadata, raw_data: Optional[List[Tuple[Any]]]
    ) -> None:
        columns = list()
        values = list()
        for col_info in metadata:
            columns.append('"' + self._get_query_table_col_name(col_info) + '"')
            values.append("?")

        self.connection.executemany(
            "INSERT INTO %s (%s) VALUES (%s)"
            % (self.query_id, ",".join(columns), ",".join(values)),
            [r for r in raw_data],
        )

        self.connection.commit()
        self.set_cache_last_updated()

    def _create_query_sql(self) -> str:
        parser = self.statement.sql_parser.parser
        query_sql = ""
        outer_columns = "*" if parser.outer_columns is None else parser.outer_columns
        outer_exprs = "" if parser.outer_exprs is None else parser.outer_exprs

        if parser.inner_columns is None:
            query_sql = "SELECT %s FROM %s %s" % (
                outer_columns,
                self.query_id,
                outer_exprs,
            )
        else:
            inner_columns = parser.inner_columns
            inner_exprs = "" if parser.inner_exprs is None else parser.inner_exprs

            query_sql = "SELECT %s FROM (SELECT %s FROM %s %s) %s" % (
                outer_columns,
                inner_columns,
                self.query_id,
                inner_exprs,
                outer_exprs,
            )

        return query_sql

    def query(self):
        query_sql = self._create_query_sql()
        with closing(self.connection.cursor()) as cursor:
            cursor.execute(query_sql)
            self.set_cache_queried_times()
            results = cursor.fetchall()
            desc = cursor.description
            return (desc, results)



# -*- coding: utf-8 -*-
import logging
from abc import ABCMeta
from .base import Base
from typing import Any, Dict, List, Optional
from .common import KeyWords, Tokens
from .util import flatten_list
from pyparsing import (
    ParseResults,
    Word,
    CaselessKeyword,
    alphanums,
    nums,
    quoted_string,
    Group,
    ZeroOrMore,
    opAssoc,
    Opt,
    delimited_list,
    one_of,
    infix_notation,
    pyparsing_common as ppc,
    Combine,
    Regex,
)

_logger = logging.getLogger(__name__)  # type: ignore


class DmlBase(Base):
    _CONSISTENT_READ, _RETURN_CONSUMED_CAPACITY = map(
        CaselessKeyword,
        [
            "ConsistentRead",
            "ReturnConsumedCapacity",
        ],
    )

    ATTR_NAME = Opt('"') + Word(alphanums + "_-") + Opt('"')
    ATTR_ARRAY_NAME = ATTR_NAME + "[" + Word(nums) + "]"

    _COLUMN_NAME = (
        KeyWords.STAR
        ^ Combine(delimited_list(ATTR_NAME ^ ATTR_ARRAY_NAME, delim=".", combine=True))
    )("column_name").set_name("column_name")

    _ALIAS_NAME = Word(alphanums + "_-")("alias_name").set_name("alias_name")

    _COLUMN = _COLUMN_NAME

    _COLUMNS = delimited_list(
        Group(
            _COLUMN
            + ZeroOrMore(Group(KeyWords.ARITHMETIC_OPERATORS + _COLUMN))(
                "column_ops"
            ).set_name("column_ops")
        )
    )("columns").set_name("columns")

    _COLUMN_RVAL = (
        ppc.real() | ppc.signed_integer() | quoted_string | _COLUMN | KeyWords.QUESTION
    )("column_rvalue").set_name("column_rvalue")

    _WHERE_CONDITION = Group(
        (_COLUMN + KeyWords.COMPARISON_OPERATORS + _COLUMN_RVAL)
        ^ (
            _COLUMN
            + KeyWords.IN
            + Group(
                "["
                + delimited_list(_COLUMN_RVAL)("in_values_list").set_name(
                    "in_values_list"
                )
                + "]"
            )
        )
        ^ (_COLUMN + KeyWords.BETWEEN + _COLUMN_RVAL + KeyWords.AND + _COLUMN_RVAL)
        ^ (_COLUMN + KeyWords.IS + Tokens.PARTIQL_DATA_TYPE)
        ^ (_COLUMN + KeyWords.IS + KeyWords.NOT + Tokens.PARTIQL_DATA_TYPE)
        ^ (
            KeyWords.FUNCTION_ON_WHERE
            + "("
            + delimited_list(quoted_string)("function_params").set_name(
                "function_params"
            )
            + ")"
        )
        ^ (
            KeyWords.FUNCTION_WITH_OP_ON_WHERE
            + "("
            + delimited_list(quoted_string)("function_params").set_name(
                "function_params"
            )
            + ")"
            + KeyWords.COMPARISON_OPERATORS
            + _COLUMN_RVAL
        )
    )("where_condition").set_name("where_condition")

    _WHERE_CONDITIONS = infix_notation(
        _WHERE_CONDITION,
        [
            (KeyWords.NOT, 1, opAssoc.RIGHT),
            (one_of("AND OR", caseless=True), 2, opAssoc.LEFT),
            ((KeyWords.BETWEEN, KeyWords.AND), 3, opAssoc.RIGHT),
        ],
        lpar="(",
        rpar=")",
    )("where_conditions").set_name("where_expression")

    _RAW_SUPPORTED_OPTIONS = ZeroOrMore(
        Group(KeyWords.ORDER_BY + _COLUMN + Tokens.ORDER_BY_VALUE)(
            "raw_supported_option"
        ).set_name("raw_supported_option")
    )("raw_supported_options").set_name("raw_supported_options")

    _OPTIONS = ZeroOrMore(
        Group(
            KeyWords.LIMIT + Opt(KeyWords.EQUALS) + Tokens.INT_VALUE
            ^ _CONSISTENT_READ + Opt(KeyWords.EQUALS) + Tokens.BOOL_VALUE
            ^ _RETURN_CONSUMED_CAPACITY
            + Opt(KeyWords.EQUALS)
            + one_of("INDEXES TOTAL NONE")
        )("option").set_name("option")
    )("options").set_name("options")

    _RETURNING_CLAUSE = Group(
        KeyWords.RETURNING + Regex(r".*")("return_content").set_name("return_content")
    )("returning_clause").set_name("returning_clause")

    def __init__(self, statement: str) -> None:
        super().__init__(statement)
        self._where_conditions = []
        self._limit = None
        self._consistent_read = False
        self._return_consumed_capacity = "NONE"

    @property
    def where_conditions(self) -> List[Optional[str]]:
        return self._where_conditions

    @property
    def limit(self) -> int:
        return self._limit

    @property
    def consistent_read(self) -> bool:
        return self._consistent_read

    @property
    def return_consumed_capacity(self) -> str:
        return self._return_consumed_capacity

    @property
    def syntax_def(self) -> None:
        return None

    def transform(self) -> Dict[str, Any]:
        return {"Statement": self._statement}

    def _construct_where_conditions(self, where_conditions: List[Any]) -> str:
        for condition in where_conditions:
            if not isinstance(condition, ParseResults):
                self._where_conditions.append(str(condition))
            else:
                function_ = condition.get("function", None)
                function_with_op_ = condition.get("function_with_op", None)
                if function_:
                    flatted_func_params = ",".join(condition["function_params"])
                    self._where_conditions.append(
                        "%s(%s)" % (function_, flatted_func_params)
                    )
                elif function_with_op_:
                    flatted_func_params = ",".join(condition["function_params"])
                    self._where_conditions.append(
                        "%s(%s) %s %s"
                        % (
                            function_with_op_,
                            flatted_func_params,
                            condition["comparison_operators"],
                            condition["column_rvalue"],
                        )
                    )
                else:
                    where_conditions_ = condition.as_list()
                    flatted_where = " ".join(
                        str(c) for c in flatten_list(where_conditions_)
                    )
                    self._where_conditions.append(flatted_where)

        return "WHERE %s" % " ".join(self.where_conditions)


class DmlFunction(metaclass=ABCMeta):
    def __init__(self, name: str, params: List[str] = None) -> None:
        self._name = name
        self._params = params

    @property
    def name(self) -> str:
        return self._name

    @property
    def params(self) -> List[str]:
        return self._params



# -*- coding: utf-8 -*-
"""
Syntax of DDB table creation with SQL:
-------------------------------------
CREATE [GLOBAL] TABLE tbl_name
    [(create_definition, ...)]
    [table_options]

create_definition: {
    col_name column_definition
    | INDEX index_name index_type
        (col_name key_type, ...)
      [index_option] ...
}

column_definition: {
    data_type [key_type]
}

key_type : {
    PARTITION KEY | SORT KEY | HASH | RANGE
}

data_type: {
    NUMERIC | STRING | BINARY
}

index_type:
    {GLOBAL | LOCAL}

index_option: {
    Projection.ProjectionType [=] {ALL | KEYS_ONLY | INCLUDE}
    | Projection.NonKeyAttributes [=] ('string', ...)
    | ProvisionedThroughput.ReadCapacityUnits [=] value
    | ProvisionedThroughput.WriteCapacityUnits [=] value
}

table_options:
    table_option [[,] table_option] ...

table_option: {
    BillingMode [=] {PROVISIONED | PAY_PER_REQUEST}
    | ProvisionedThroughput.ReadCapacityUnits [=] value
    | ProvisionedThroughput.WriteCapacityUnits [=] value
    | StreamSpecification.StreamEnabled [=] {True | False}
    | StreamSpecification.StreamViewType [=] {NEW_IMAGE | OLD_IMAGE | NEW_AND_OLD_IMAGES | KEYS_ONLY}
    | SSESpecification.Enabled [=] {True | False}
    | SSESpecification.SSEType [=] {AES256 | KMS}
    | SSESpecification.KMSMasterKeyId [=] value
    | TableClass [=] {STANDARD | STANDARD_INFREQUENT_ACCESS}
    | Tags [=] (tags)
    | ReplicationGroup [=] ('string', ...)
}

tags:
    key:value, ...


Sample SQL of Creating Table:
-----------------------------
CREATE TABLE Issues (
    IssueId numeric PARTITION KEY,
    Title string SORT KEY,
    CreateDate string,
    DueDate string,
    INDEX CreateDateIndex GLOBAL (
        CreateDate PARTITION KEY,
        IssueId SORT KEY
    )
        Projection.ProjectionType INCLUDE
        Projection.NonKeyAttributes (Description, Status)
        ProvisionedThroughput.ReadCapacityUnits 1
        ProvisionedThroughput.WriteCapacityUnits 1,
    INDEX DueDateIndex GLOBAL (
        DueDate PARTITION KEY
    )
        Projection.ProjectionType ALL
        ProvisionedThroughput.ReadCapacityUnits 1
        ProvisionedThroughput.WriteCapacityUnits 1,
)
BillingMode PROVISIONED
ProvisionedThroughput.ReadCapacityUnits 1
ProvisionedThroughput.WriteCapacityUnits 1
StreamSpecification.StreamEnabled False
SSESpecification.Enabled False
TableClass STANDARD
Tags (name:Issue, usage:test)

CREATE GLOBAL TABLE Issues
    ReplicationGroup (us-east-1, us-west-2)
"""

import logging
from .common import KeyWords, Tokens
from .ddl_sql import DdlBase
from pyparsing import (
    Opt,
    Group,
    ZeroOrMore,
    OneOrMore,
    delimited_list,
    Forward,
)
from typing import Any, Dict

_logger = logging.getLogger(__name__)  # type: ignore


class DdlCreate(DdlBase):
    _ATTRIBUTE = Group(Tokens.ATTRIBUTE_NAME + Tokens.DATA_TYPE + Tokens.OPT_KEY_TYPE)(
        "attribute"
    ).set_name("attribute")

    _CREATE_TABLE_STATEMENT = (
        KeyWords.CREATE
        + Opt(KeyWords.GLOBAL)("global").set_name("global")
        + KeyWords.TABLE
        + Tokens.TABLE_NAME
        + Opt(
            KeyWords.LPAR
            + delimited_list(OneOrMore(_ATTRIBUTE))("attributes").set_name("attributes")
            + delimited_list(ZeroOrMore(DdlBase._INDEX))("indices").set_name("indices")
            + KeyWords.RPAR
        )
        + DdlBase._TABLE_OPTIONS
    )("create_statement").set_name("create_statement")

    _DDL_CREATE_EXPR = Forward()
    _DDL_CREATE_EXPR <<= _CREATE_TABLE_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    @property
    def syntax_def(self) -> Forward:
        return DdlCreate._DDL_CREATE_EXPR

    def transform(self) -> Dict[str, Any]:
        if self.root_parse_results is None:
            raise ValueError("Statement was not parsed yet")

        request = dict()
        is_global_table = (
            True if self.root_parse_results.get("global", None) is not None else False
        )
        table_name_ = self.root_parse_results["table"]

        if is_global_table:
            request.update({"GlobalTableName": table_name_})
        else:
            attr_def_ = list()
            key_schema_ = list()

            for attr in self.root_parse_results["attributes"]:
                attr_name = attr["attribute_name"]
                data_type = attr["data_type"]
                attr_def_.append(self._construct_attr_def(attr_name, data_type))

                key_type = attr["key_type"]
                if key_type != "":
                    key_schema_.append(self._construct_key_schema(attr_name, key_type))
            request.update({"AttributeDefinitions": attr_def_})

            request.update({"TableName": table_name_})
            request.update({"KeySchema": key_schema_})

            lsis_ = None
            gsis_ = None
            indices = self.root_parse_results["indices"]
            if indices is not None:
                for index in indices:
                    index_type = index["index_type"]

                    if index_type == "GLOBAL":
                        if gsis_ is None:
                            gsis_ = list()
                        gsis_.append(self._construct_index(index))
                    elif index_type == "LOCAL":
                        if lsis_ is None:
                            lsis_ = list()
                        lsis_.append(self._construct_index(index))
                    else:
                        raise LookupError("Index type is invalid")

            if lsis_ is not None:
                request.update({"LocalSecondaryIndexes": lsis_})
            if gsis_ is not None:
                request.update({"GlobalSecondaryIndexes": gsis_})

        table_options_ = self.root_parse_results["table_options"]
        if table_options_ is not None:
            converted_options = self._construct_options(table_options_)
            if converted_options is not None:
                request.update(converted_options)

        return request



# -*- coding: utf-8 -*-
"""
Syntax of PartiQL
https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/ql-reference.delete.html
------------------
DELETE FROM table
WHERE condition [RETURNING returnvalues]
<returnvalues>  ::= ALL OLD *

Sample SQL of Deleting:
------------------------
DELETE FROM "Music" WHERE "Artist" = 'Acme Band' AND "SongTitle" = 'PartiQL Rocks'

DELETE FROM "Music" WHERE "Artist" = 'Acme Band' AND "SongTitle" = 'PartiQL Rocks' RETURNING ALL OLD *
"""
import logging
from .dml_sql import DmlBase
from .common import KeyWords, Tokens
from pyparsing import (
    Forward,
    Opt,
)
from typing import Any, Dict

_logger = logging.getLogger(__name__)  # type: ignore


class DmlDelete(DmlBase):

    _DELETE_STATEMENT = (
        KeyWords.DELETE
        + KeyWords.FROM
        + Tokens.TABLE_NAME
        + KeyWords.WHERE
        + DmlBase._WHERE_CONDITIONS
        + Opt(DmlBase._RETURNING_CLAUSE)
    )("delete_statement").set_name("delete_statement")

    _DML_DELETE_EXPR = Forward()
    _DML_DELETE_EXPR <<= _DELETE_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    @property
    def syntax_def(self) -> Forward:
        return DmlDelete._DML_DELETE_EXPR

    def transform(self) -> Dict[str, Any]:
        table_name_ = self.root_parse_results["table"]

        table_ = '"%s"' % table_name_

        # Build the statement
        statement_ = "DELETE FROM {table} {where_conditions}"

        where_conditions = self.root_parse_results.get("where_conditions", None)
        if where_conditions is not None:
            where_conditions_ = self._construct_where_conditions(where_conditions)
        else:
            where_conditions_ = ""

        statement_ = statement_.format(
            table=table_, where_conditions=str(where_conditions_)
        )

        # Add RETURNING clause if present
        if "returning_clause" in self.root_parse_results:
            returning = self.root_parse_results["returning_clause"]
            return_content = returning.get("return_content", "").strip()
            statement_ += " RETURNING %s" % return_content

        request = {"Statement": statement_.strip()}
        return request



# -*- coding: utf-8 -*-
"""
Syntax of PartiQL
https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/ql-reference.update.html
------------------
UPDATE  table
[SET | REMOVE]  path  [=  data] […]
WHERE condition [RETURNING returnvalues]
<returnvalues>  ::= [ALL OLD | MODIFIED OLD | ALL NEW | MODIFIED NEW] *

Sample SQL of Updating:
------------------------
UPDATE "Music"
SET AwardsWon=1
SET AwardDetail={'Grammys':[2020, 2018]}
WHERE Artist='Acme Band' AND SongTitle='PartiQL Rocks'
RETURNING ALL OLD *

UPDATE "Music"
REMOVE AwardDetail.Grammys[2]
WHERE Artist='Acme Band' AND SongTitle='PartiQL Rocks'
"""
import logging
import re
from .dml_sql import DmlBase
from .common import KeyWords, Tokens
from pyparsing import Forward, Group, OneOrMore, Opt, Regex
from typing import Any, Dict

_logger = logging.getLogger(__name__)  # type: ignore


class DmlUpdate(DmlBase):

    # Define SET operation: SET followed by content until next SET/REMOVE/WHERE
    _SET_OPERATION = Group(
        KeyWords.SET
        + Regex(r".*?(?=\s+(?:SET|REMOVE|WHERE))", re.IGNORECASE | re.DOTALL)(
            "set_content"
        ).set_name("set_content")
    )("set_op")

    # Define REMOVE operation: REMOVE followed by content until next SET/REMOVE/WHERE
    _REMOVE_OPERATION = Group(
        KeyWords.REMOVE
        + Regex(r".*?(?=\s+(?:SET|REMOVE|WHERE))", re.IGNORECASE | re.DOTALL)(
            "remove_content"
        ).set_name("remove_content")
    )("remove_op")

    # Multiple SET or REMOVE operations
    _OPERATIONS = Group(OneOrMore(_SET_OPERATION | _REMOVE_OPERATION))("operations")

    _UPDATE_STATEMENT = (
        KeyWords.UPDATE
        + Tokens.TABLE_NAME
        + _OPERATIONS
        + KeyWords.WHERE
        + DmlBase._WHERE_CONDITIONS
        + Opt(DmlBase._RETURNING_CLAUSE)
    )("update_statement").set_name("update_statement")

    _DML_UPDATE_EXPR = Forward()
    _DML_UPDATE_EXPR <<= _UPDATE_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    @property
    def syntax_def(self) -> Forward:
        return DmlUpdate._DML_UPDATE_EXPR

    def transform(self) -> Dict[str, Any]:
        table_name_ = self.root_parse_results["table"]
        operations_ = self.root_parse_results["operations"]

        table_ = '"%s"' % table_name_

        # Build the operations part from multiple SET/REMOVE operations
        operations_parts = []
        for op in operations_:
            if "set_op" == op.get_name():
                operations_parts.append("SET %s" % op["set_content"].strip())
            elif "remove_op" == op.get_name():
                operations_parts.append("REMOVE %s" % op["remove_content"].strip())

        operations_str = " ".join(operations_parts)

        where_conditions = self.root_parse_results.get("where_conditions", None)
        if where_conditions is not None:
            where_conditions_ = self._construct_where_conditions(where_conditions)
        else:
            where_conditions_ = ""

        # Build the statement
        statement_ = "UPDATE {table} {operations} {where_conditions}"

        statement_ = statement_.format(
            table=table_,
            operations=operations_str,
            where_conditions=where_conditions_,
        )

        # Add RETURNING clause if present
        if "returning_clause" in self.root_parse_results:
            returning = self.root_parse_results["returning_clause"]
            return_content = returning.get("return_content", "").strip()
            statement_ += " RETURNING %s" % return_content

        request = {"Statement": statement_.strip()}
        return request



# -*- coding: utf-8 -*-
"""
Syntax of PartiQL
https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/ql-reference.select.html
------------------
SELECT expression  [, ...]
FROM table[.index]
[ WHERE condition ] [ [ORDER BY key [DESC|ASC]] , ...]

Extension of PartiQL:
---------------------
Add Limit, ConsistentRead, ReturnConsumedCapacity options to the tailing
SELECT [expression | function(expression)] [alias]  [, ...]
FROM table[.index]
[ WHERE condition ]
[ [ORDER BY key [DESC|ASC]]
  [Limit value]
  [ConsistentRead {True|False}]
  [ReturnConsumedCapacity {INDEXES|TOTAL|NONE}]
...]

Sample SQL of Selecting:
------------------------
SELECT *
FROM "Issues"."CreateDateIndex"
WHERE IssueId IN [100, 300, 234]
AND Title = 'some title'
AND Content[0] >= 100
ORDER BY IssueId DESC
LIMIT 10
ConsistentRead False
ReturnConsumedCapacity NONE
"""
import logging
import re
from abc import ABCMeta
from collections import OrderedDict
from .dml_sql import DmlBase, DmlFunction
from .common import KeyWords, Tokens
from pyparsing import Opt, Forward, Group, ZeroOrMore, delimited_list, Regex
from typing import Any, Dict, List, Optional

_logger = logging.getLogger(__name__)  # type: ignore


class DmlSelectColumn(metaclass=ABCMeta):
    def __init__(
        self,
        request_name: str,
        alias: str = None,
        result_name: str = None,
        function: DmlFunction = None,
    ) -> None:
        self._request_name = request_name
        self._alias = alias
        self._result_name = result_name
        self._function = function

    @property
    def request_name(self) -> str:
        return self._request_name

    @property
    def alias(self) -> str:
        return self._alias

    @alias.setter
    def alias(self, value: str) -> None:
        self._alias = value

    @property
    def result_name(self) -> str:
        if self._result_name is not None:
            return self._result_name

        if self.request_name is not None:
            self._result_name = self.request_name.split(".")[-1]
            self._result_name = self._result_name.replace('"', "")

        return self._result_name

    @property
    def function(self) -> DmlFunction:
        return self._function

    def __str__(self):
        return "(request_name: %s, alias: %s, result_name: %s, function: %s)" % (
            self.request_name,
            self.alias,
            self.result_name,
            self.function.name if self.function is not None else None,
        )

    __repr__ = __str__


class DmlSelect(DmlBase):
    _FUNCTION_PARAMS = ZeroOrMore(
        KeyWords.COMMA
        + Tokens.QUOTED_STRING("function_param").set_name("function_param")
        | KeyWords.COMMA + Tokens.INT_VALUE("function_param").set_name("function_param")
    )("function_params").set_name("function_params")

    _COLUMN_WITH_FUNCTION = (
        KeyWords.FUNCTION_ON_COLUMN
        + KeyWords.LPAR
        + DmlBase._COLUMN_NAME
        + _FUNCTION_PARAMS
        + KeyWords.RPAR
    )

    _REQUEST_COLUMN = _COLUMN_WITH_FUNCTION | DmlBase._COLUMN_NAME

    _ALIAS = (
        Opt(KeyWords.SUPPRESS_QUOTE)
        + DmlBase._ALIAS_NAME
        + Opt(KeyWords.SUPPRESS_QUOTE)
    )

    _REQUEST_COLUMNS = delimited_list(
        Group(
            Opt(KeyWords.LPAR)
            + _REQUEST_COLUMN
            + ZeroOrMore(Group(KeyWords.ARITHMETIC_OPERATORS + _REQUEST_COLUMN))(
                "column_ops"
            ).set_name("column_ops")
            + Opt(KeyWords.RPAR)
            + Opt(~Regex("FROM", flags=re.IGNORECASE) + _ALIAS)
        )
    )("columns").set_name("columns")

    _SELECT_STATEMENT = (
        KeyWords.SELECT
        + _REQUEST_COLUMNS
        + KeyWords.FROM
        + Tokens.TABLE_NAME
        + Opt(KeyWords.DOT + Tokens.INDEX_NAME)
        + Opt(KeyWords.WHERE + DmlBase._WHERE_CONDITIONS)
        + Opt(DmlBase._RAW_SUPPORTED_OPTIONS)
        + Opt(DmlBase._OPTIONS)
    )("select_statement").set_name("select_statement")

    _DML_SELECT_EXPR = Forward()
    _DML_SELECT_EXPR <<= _SELECT_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)
        self._columns = []
        self._is_star_column = False

    @property
    def columns(self) -> List[Optional[DmlSelectColumn]]:
        return self._columns

    @property
    def is_star_column(self) -> bool:
        return self._is_star_column

    @property
    def syntax_def(self) -> Forward:
        return DmlSelect._DML_SELECT_EXPR

    def transform(self) -> Dict[str, Any]:
        table_name_ = self.root_parse_results["table"]
        index_name_ = self.root_parse_results.get("index_name", None)

        table_ = None
        if index_name_ is None:
            table_ = '"%s"' % table_name_
        else:
            table_ = '"%s"."%s"' % (table_name_, index_name_)

        columns_ = self._construct_columns(self.root_parse_results["columns"])

        where_conditions = self.root_parse_results.get("where_conditions", None)
        if where_conditions is not None:
            where_conditions_ = self._construct_where_conditions(where_conditions)
        else:
            where_conditions_ = ""

        raw_supported_options = self.root_parse_results.get("raw_supported_options", [])
        raw_supported_options_ = self._construct_raw_options(raw_supported_options)
        if raw_supported_options_ is not None:
            raw_supported_options_ = " ".join(raw_supported_options_)
        else:
            raw_supported_options_ = ""

        statement_ = "SELECT {columns} FROM {table} {where_conditions} {options}"
        statement_ = statement_.format(
            columns=columns_,
            table=table_,
            where_conditions=where_conditions_,
            options=raw_supported_options_,
        )
        request = {"Statement": statement_.strip()}

        options = self.root_parse_results.get("options", [])
        options_ = self._construct_options(options)
        if options_ is not None:
            request.update(options_)

        return request

    def _construct_columns(self, columns: List[Any]) -> str:
        columns_ = (
            OrderedDict()
        )  # Create a ordered dict to avoid duplicate column names
        for column in columns:
            if column["column_name"] == "*":
                self._is_star_column = True
                self._columns.clear()
                return "*"

            column_ = []
            column_.append(column["column_name"])

            for rcolumn in column["column_ops"]:
                column_.append(rcolumn["arithmetic_operators"])
                column_.append(rcolumn["column_name"])

            column__ = "".join(column_)
            columns_[column__] = (
                "_"  # Add a dummy value to avoid duplicate column names
            )

            function_ = column.get("function", None)
            column_function_ = None
            if function_:
                func_params_ = column["function_params"]
                if len(func_params_) > 0:
                    column_function_ = DmlFunction(
                        function_, params=[p for p in func_params_]
                    )
                else:
                    column_function_ = DmlFunction(function_)

            alias_ = column.get("alias_name", None)

            self._columns.append(
                DmlSelectColumn(column__, function=column_function_, alias=alias_)
            )
        columns_ = ",".join(columns_.keys())
        return columns_

    def _construct_raw_options(self, options: List[Any]) -> Optional[List[Any]]:
        converted_ = None
        for option in options:
            if converted_ is None:
                converted_ = []

            converted_.append(" ".join(str(o) for o in option))

        return converted_

    def _construct_options(self, options: List[Any]) -> Optional[Dict[str, Any]]:
        converted_ = None
        for option in options:
            if converted_ is None:
                converted_ = {}
            option_name = option[0]
            option_value = option[1]

            if option_name == "Limit":
                self._limit = option_value
                continue
            elif option_name == "ConsistentRead":
                self._consistent_read = option_value
            elif option_name == "ReturnConsumedCapacity":
                self._return_consumed_capacity = option_value

            converted_.update({option_name: option_value})

        return converted_



# -*- coding: utf-8 -*-
import logging
from abc import ABCMeta

from .common import QueryType, get_query_type
from .base import Base
from .ddl_create import DdlCreate
from .ddl_alter import DdlAlter
from .ddl_drop import DdlDrop
from .dml_insert import DmlInsert
from .dml_select import DmlSelect
from .dml_update import DmlUpdate
from .dml_delete import DmlDelete
from .util_sql import UtilListTables, UtilDescTable
from typing import Any, Dict, Tuple

_logger = logging.getLogger(__name__)  # type: ignore


class SQLParser(metaclass=ABCMeta):
    def __init__(self, statement: str = "", parser_class: Base = None) -> None:
        self._statement = statement
        self._query_type = None
        self._query_category = None
        self._operation_type = None
        self._parser_class = parser_class
        self._parser = None

    @property
    def statement(self) -> str:
        return self._statement

    @property
    def query_type(self) -> Tuple[str, str, str]:
        if self._statement is None:
            raise ValueError("Statement is not specified")

        if self._query_type is not None:
            return self._query_type

        self._query_type = get_query_type(self._statement)
        return self._query_type

    @property
    def query_category(self) -> str:
        if self._query_category is not None:
            return self._query_category

        self._query_category = (
            self.query_type[0] if self.query_type is not None else None
        )
        return self._query_category

    @property
    def operation_type(self) -> str:
        if self._operation_type is not None:
            return self._operation_type

        self._operation_type = (
            self.query_type[1] if self.query_type is not None else None
        )
        return self._operation_type

    @property
    def parser(self) -> Base:
        if self._parser is None:
            if self._parser_class is None:
                self._parser = self._get_parse_class()(self.statement)
            else:
                self._parser = self._parser_class(self.statement)

        return self._parser

    def _get_parse_class(self) -> Base:
        _parse_class = None
        if (
            self.query_type == QueryType.CREATE
            or self.query_type == QueryType.CREATE_GLOBAL
        ):
            _parse_class = DdlCreate
        elif self.query_type == QueryType.ALTER:
            _parse_class = DdlAlter
        elif (
            self.query_type == QueryType.DROP
            or self.query_type == QueryType.DROP_GLOBAL
        ):
            _parse_class = DdlDrop
        elif self.query_type == QueryType.SELECT:
            _parse_class = DmlSelect
        elif self.query_type == QueryType.INSERT:
            _parse_class = DmlInsert
        elif self.query_type == QueryType.UPDATE:
            _parse_class = DmlUpdate
        elif self.query_type == QueryType.DELETE:
            _parse_class = DmlDelete
        elif (
            self.query_type == QueryType.LIST
            or self.query_type == QueryType.LIST_GLOBAL
        ):
            _parse_class = UtilListTables
        elif (
            self.query_type == QueryType.DESC
            or self.query_type == QueryType.DESC_GLOBAL
        ):
            _parse_class = UtilDescTable

        return _parse_class

    def transform(self) -> Dict[str, Any]:
        assert self.parser is not None, "Parser not found."
        return self.parser.transform()



# -*- coding: utf-8 -*-
import logging
from .base import Base
from .common import KeyWords, Tokens
from pyparsing import Opt, Group, CaselessKeyword, ZeroOrMore, Forward
from typing import Any, Dict

_logger = logging.getLogger(__name__)  # type: ignore


class UtilBase(Base):
    def __init__(self, statement: str) -> None:
        super().__init__(statement)


"""
Syntax of list DDB tables:
{LIST | SHOW} [GLOBAL] TABLES
    [Limit [=] value]
    [RegionName [=] 'string']

Sample SQL of Listing Tables:
-----------------------------
LIST TABLES
SHOW TABLES Limit 10

LIST GLOBAL TABLES
    Limit 10
    RegionName us-west-1
"""


class UtilListTables(UtilBase):
    _REGION_NAME = CaselessKeyword("RegionName")

    _LIST_TABLES_STATEMENT = (
        (KeyWords.LIST | KeyWords.SHOW)
        + Opt(KeyWords.GLOBAL)("global").set_name("global")
        + KeyWords.TABLES
        + ZeroOrMore(
            Group(
                KeyWords.LIMIT
                + Opt(KeyWords.EQUALS)
                + Tokens.INT_VALUE("limit").set_name("limit")
                ^ _REGION_NAME + Opt(KeyWords.EQUALS) + Tokens.REGION_NAME
            )("option").set_name("option")
        )("options").set_name("options")
    )("list_tables_statement").set_name("list_tables_statement")

    _UTIL_LIST_TABLES_EXPR = Forward()
    _UTIL_LIST_TABLES_EXPR <<= _LIST_TABLES_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)
        self._limit = None

    @property
    def limit(self) -> int:
        return self._limit

    @property
    def syntax_def(self) -> Forward:
        return UtilListTables._UTIL_LIST_TABLES_EXPR

    def transform(self) -> Dict[str, Any]:
        if self.root_parse_results is None:
            raise ValueError("Statement was not parsed yet")

        request = dict()
        is_global_table = (
            True if self.root_parse_results.get("global", None) is not None else False
        )
        options = self.root_parse_results["options"]

        for option in options:
            limit_value = option.get("limit", None)
            region_name = option.get("region_name", None)

            if limit_value:
                self._limit = limit_value
                request.update({"Limit": limit_value})

            if is_global_table and region_name:
                request.update({"RegionName": region_name})

        return request


"""
Syntax of describe DDB table:
{DESC | DESCRIBE} [GLOBAL] tbl_name

Sample SQL of Describing Table:
-----------------------------
DESC Issues
DESCRIBE Issues

DESC GLOBAL Issues
"""


class UtilDescTable(UtilBase):
    _DESC_TABLE_STATEMENT = (
        (KeyWords.DESC | KeyWords.DESCRIBE)
        + Opt(KeyWords.GLOBAL)("global").set_name("global")
        + Tokens.TABLE_NAME
    )("desc_table_statement").set_name("desc_table_statement")

    _DESC_TABLE_EXPR = Forward()
    _DESC_TABLE_EXPR <<= _DESC_TABLE_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    @property
    def syntax_def(self) -> Forward:
        return UtilDescTable._DESC_TABLE_EXPR

    def transform(self) -> Dict[str, Any]:
        if self.root_parse_results is None:
            raise ValueError("Statement was not parsed yet")

        request = dict()
        table_name = self.root_parse_results["table"]
        is_global_table = (
            True if self.root_parse_results.get("global", None) is not None else False
        )

        if is_global_table:
            request.update({"GlobalTableName": table_name})
        else:
            request.update({"TableName": table_name})

        return request



from sqlalchemy.dialects import registry

registry.register(
    "dynamodb", "pydynamodb.sqlalchemy_dynamodb.pydynamodb", "DynamoDBDialect"
)
registry.register(
    "dynamodb.rest", "pydynamodb.sqlalchemy_dynamodb.pydynamodb", "DynamoDBRestDialect"
)



# -*- coding: utf-8 -*-
import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast, TypeVar, Tuple

from .converter import Converter
from .common import BaseCursor, CursorIterator
from .model import Statement, Statements
from .result_set import DynamoDBResultSet, DynamoDBDictResultSet
from .error import NotSupportedError, ProgrammingError
from .util import RetryConfig, synchronized

if TYPE_CHECKING:
    from .connection import Connection

_logger = logging.getLogger(__name__)  # type: ignore
_T = TypeVar("_T", bound="Cursor")


class Cursor(BaseCursor, CursorIterator):
    NO_RESULT_SET = "No result set."

    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        retry_config: RetryConfig,
        **kwargs
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            retry_config=retry_config,
            **kwargs,
        )
        self._result_set: Optional[DynamoDBResultSet] = None
        self._result_set_class = DynamoDBResultSet
        self._statements: Statements = Statements()
        self._transaction_statements: Statements = Statements()
        self._is_pooling: bool = False
        self._kwargs = kwargs

    @property
    def result_set(self) -> Optional[CursorIterator]:
        return self._result_set

    @result_set.setter
    def result_set(self, val) -> None:
        self._result_set = val

    @property
    def has_result_set(self) -> bool:
        return self._result_set is not None

    @property
    def result_set_class(self) -> Optional[CursorIterator]:
        return self._result_set_class

    @result_set_class.setter
    def result_set_class(self, val) -> None:
        self._result_set_class = val

    @property
    def rowcount(self) -> int:
        return self._result_set.rowcount if self._result_set else -1

    @property
    def rownumber(self) -> Optional[int]:
        return self._result_set.rownumber if self._result_set else None

    @property
    def description(
        self,
    ) -> Optional[List[Tuple[str, str, None, None, None, None, None]]]:
        return self._result_set.description

    @property
    def errors(self) -> List[Dict[str, str]]:
        return self._result_set.errors

    @synchronized
    def execute(
        self: _T, operation: str, parameters: Optional[List[Dict[str, Any]]] = None
    ) -> _T:
        statement = Statement(operation)
        return self.execute_statement(statement, parameters)

    @synchronized
    def execute_statement(
        self: _T,
        statement: Statement,
        parameters: Optional[List[Dict[str, Any]]] = None,
    ) -> _T:
        try:
            if parameters:
                statement.api_request.update(
                    {
                        "Parameters": [
                            self._converter.serialize(parameter)
                            for parameter in parameters
                        ],
                    }
                )
            if self.connection.in_transaction:
                self._transaction_statements.append(statement)
            else:
                self._statements.append(statement)

            if not self._is_pooling and self.connection.autocommit:
                self.flush()

        except Exception as e:
            if self.connection.in_transaction:
                self.connection.in_transaction = False
                self.connection.autocommit = True
            raise e

        return self

    @synchronized
    def executemany(
        self,
        operation: str,
        seq_of_parameters: List[Optional[Dict[str, Any]]],
    ) -> None:
        self._is_pooling = True
        for i, parameters in enumerate(seq_of_parameters):
            if i == len(seq_of_parameters) - 1:
                self._is_pooling = False
            self.execute(operation, parameters)

    @synchronized
    def execute_transaction(self) -> None:
        if len(self._transaction_statements) > 0:
            try:
                self._reset_state()
                self._result_set = self._result_set_class(
                    self._connection,
                    self._converter,
                    deepcopy(self._transaction_statements),
                    self.arraysize,
                    self._retry_config,
                    is_transaction=True,
                    **self._kwargs,
                )
            finally:
                self._transaction_statements.clear()

    @synchronized
    def flush(self) -> None:
        if len(self._statements) > 0:
            self._reset_state()
            try:
                self._result_set = self._result_set_class(
                    self._connection,
                    self._converter,
                    deepcopy(self._statements),
                    self.arraysize,
                    self._retry_config,
                    is_transaction=False,
                    **self._kwargs,
                )
            finally:
                self._statements.clear()
                self._is_pooling = False
                self.connection.autocommit = True

    def fetchone(
        self,
    ) -> Optional[Dict[Any, Optional[Any]]]:
        if not self.has_result_set:
            raise ProgrammingError(Cursor.NO_RESULT_SET)
        result_set = cast(DynamoDBResultSet, self._result_set)
        return result_set.fetchone()

    def fetchmany(self, size: int = None) -> Optional[Dict[Any, Optional[Any]]]:
        if not self.has_result_set:
            raise ProgrammingError(Cursor.NO_RESULT_SET)
        result_set = cast(DynamoDBResultSet, self._result_set)
        return result_set.fetchmany(size)

    def fetchall(
        self,
    ) -> Optional[Dict[Any, Optional[Any]]]:
        if not self.has_result_set:
            raise ProgrammingError(Cursor.NO_RESULT_SET)
        result_set = cast(DynamoDBResultSet, self._result_set)
        return result_set.fetchall()

    def cancel(self) -> None:
        raise NotSupportedError

    def close(self) -> None:
        self._reset_state()
        self._statements = None
        self._transaction_statements = None

    def _reset_state(self) -> None:
        if self.result_set and not self.result_set.is_closed:
            self.result_set.close()
        self.result_set = None  # type: ignore


class DictCursor(Cursor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        retry_config: RetryConfig,
        **kwargs
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            retry_config=retry_config,
            **kwargs,
        )
        self._result_set: Optional[DynamoDBDictResultSet] = None
        self._result_set_class = DynamoDBDictResultSet



# -*- coding: utf-8 -*-
import logging
import json
from datetime import datetime
from abc import ABCMeta, abstractmethod
from collections import deque
from typing import TYPE_CHECKING, Any, Deque, Dict, List, Optional, Tuple, cast

from .sql.common import DdbDataTypes, DataTypes, Functions, QueryType
from .sql.parser import SQLParser
from .converter import Converter
from .model import Metadata, ColumnInfo, Statements
from .error import OperationalError, DataError
from .util import RetryConfig, retry_api_call

if TYPE_CHECKING:
    from .connection import Connection

_logger = logging.getLogger(__name__)  # type: ignore


class BaseExecutor(metaclass=ABCMeta):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        self._connection: Optional["Connection"] = connection
        self._converter = converter
        self._statements = statements
        assert (
            self._statements and len(self._statements) > 0
        ), "Required statements not found."

        self._retry_config = retry_config
        self._next_token: Optional[str] = None
        self._metadata: Metadata = Metadata()
        self._is_predef_metadata: bool = False
        self._rows: Deque[Tuple[Any]] = deque()
        self._errors: List[Dict[str, str]] = list()
        self._kwargs = kwargs
        self.pre_execute()

    @property
    def connection(self) -> "Connection":
        return self._connection

    @property
    def next_token(self) -> Optional[str]:
        return self._next_token

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @property
    def rows(self) -> Deque[Dict[str, Optional[Any]]]:
        return self._rows

    @property
    def errors(self) -> List[Dict[str, str]]:
        return self._errors

    def pre_execute(self) -> None:
        self.execute()

    @abstractmethod
    def execute(self) -> None:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def process_rows(self, response: Dict[str, Any]) -> None:
        raise NotImplementedError  # pragma: no cover

    def post_execute(self):
        pass  # pragma: no cover

    def _dispatch_api_call(
        self, api_call_func, request: Dict[str, Any]
    ) -> Dict[str, Any]:
        try:
            response = retry_api_call(
                api_call_func,
                config=self._retry_config,
                logger=_logger,
                **request,
            )
        except Exception as e:
            _logger.exception("Failed to execute statement.")
            raise OperationalError(*e.args) from e
        else:
            return cast(Dict[str, Any], response)

    def close(self) -> None:
        self._metadata = None
        self._rows = self._rows.clear()
        self._errors = self._errors.clear()
        self._next_token = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def dispatch_executor(
    connection: "Connection",
    converter: Converter,
    statements: Statements,
    retry_config: RetryConfig,
    is_transaction: bool = False,
    executor_class: BaseExecutor = None,
    **kwargs,
) -> Optional[BaseExecutor]:
    if statements is None or len(statements) == 0:
        return None

    _executor_class = (
        _get_executor_class(statements, is_transaction)
        if executor_class is None
        else executor_class
    )

    return _executor_class(connection, converter, statements, retry_config, **kwargs)


def _get_executor_class(statements: Statements, is_transaction: bool):
    if statements.query_type[0] == "DML":
        return _get_dml_executor_class(statements, is_transaction)
    elif statements.query_type == QueryType.CREATE:
        return DdlCreateExecutor
    elif statements.query_type == QueryType.ALTER:
        return DdlAlterExecutor
    elif statements.query_type == QueryType.DROP:
        return DdlDropExecutor
    elif statements.query_type == QueryType.LIST:
        return UtilListTablesExecutor
    elif statements.query_type == QueryType.DESC:
        return UtilDescTableExecutor
    elif statements.query_type == QueryType.CREATE_GLOBAL:
        return DdlCreateGlobalExecutor
    elif statements.query_type == QueryType.DROP_GLOBAL:
        return DdlDropGlobalExecutor
    elif statements.query_type == QueryType.LIST_GLOBAL:
        return UtilListGlobalTablesExecutor
    elif statements.query_type == QueryType.DESC_GLOBAL:
        return UtilDescGlobalTableExecutor
    else:
        raise LookupError(
            "Not support executor for query type: %s" % str(statements.query_type)
        )


def _get_dml_executor_class(statements: Statements, is_transaction: bool):
    if len(statements) > 1:
        if is_transaction:
            return DmlTransactionExecutor
        else:
            return DmlBatchExecutor
    else:
        return DmlStatementExecutor


class DmlStatementExecutor(BaseExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        self._statement = statements[0]
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        request = self._statement.api_request

        if self.next_token:
            request.update({"NextToken": self.next_token})

        response = self._dispatch_api_call(
            self.connection.client.execute_statement, request
        )

        try:
            self.process_rows(response)
        finally:
            self.post_execute()

    def process_rows(self, response: Dict[str, Any]) -> None:
        self._process_predef_metadata(self._statement.sql_parser)

        rows = response.get("Items", None)
        if rows is None:
            raise DataError("KeyError `Items`")

        processed_rows = list()
        for row in rows:
            if self._is_predef_metadata:
                row_ = self._process_predef_row_item(row)
            else:
                row_ = self._process_undef_row_item(row)
            processed_rows.append(row_)

        self._rows.extend(processed_rows)
        self._next_token = response.get("NextToken", None)

    def _process_undef_row_item(self, row) -> Optional[Tuple]:
        row_ = dict()
        for col, val in row.items():
            type_ = next(iter(val.keys()))
            val_ = self._converter.deserialize(val)
            col_index = self._process_metadata(col, type_)
            row_[col_index] = val_

        row__ = list()
        for i in range(len(self._metadata)):
            row__.append(None)
        for i, v in row_.items():
            row__[i] = v

        return tuple(row__)

    def _process_metadata(self, col_name: str, type: str) -> int:
        type_ = DataTypes.STRING
        if type == DdbDataTypes.NULL or type == DdbDataTypes.BOOLEAN:
            type_ = DataTypes.BOOL
        elif type == DdbDataTypes.NUMBER:
            type_ = DataTypes.NUMBER

        if col_name not in self._metadata:
            self._metadata.update(ColumnInfo(col_name, col_name, type_code=type_))

        return self._metadata.index(col_name)

    def _process_predef_row_item(self, row) -> Optional[Tuple]:
        row_ = [None for _ in range(len(self.metadata))]
        for col, val in row.items():
            col_info = self.metadata.get(col, None)
            if col_info:
                if col_info.function:
                    val_ = self._converter.deserialize(
                        val,
                        function=col_info.function.name,
                        function_params=col_info.function.params,
                    )
                else:
                    val_ = self._converter.deserialize(val)
                index = self.metadata.index(col)
                row_[index] = val_

        return tuple(row_)

    def _process_predef_metadata(self, sql_parser: SQLParser) -> None:
        if sql_parser.query_type == QueryType.SELECT:
            if not sql_parser.parser.is_star_column:
                self._collect_column_info(sql_parser.parser.columns)
                self._is_predef_metadata = True
        else:
            self._is_predef_metadata = False

    def _collect_column_info(self, columns: List[Optional[Any]]) -> None:
        for column in columns:
            type_ = DataTypes.STRING
            if column.function is not None:
                function_name = column.function.name
                if (
                    function_name
                    in Functions.SUPPORTED_FUNTIONS[Functions.TYPE_CONVERSION]
                ):
                    type_ = function_name

            self._metadata.update(
                ColumnInfo(
                    column.result_name,
                    column.request_name,
                    column.alias,
                    function=column.function,
                    type_code=type_,
                )
            )


class DmlStatementDictExecutor(DmlStatementExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def process_rows(self, response: Dict[str, Any]) -> None:
        rows = response.get("Items", None)
        if rows is None:
            raise DataError("KeyError `Items`")

        processed_rows = list()
        for row in rows:
            row_ = self._process_row_item(row)
            processed_rows.append(row_)

        self._rows.extend(processed_rows)
        self._next_token = response.get("NextToken", None)

    def _process_row_item(self, row) -> Optional[Dict[str, Any]]:
        row_ = dict()
        for col, val in row.items():
            val_ = self._converter.deserialize(val)
            row_[col] = val_

        return row_


class DmlBatchExecutor(DmlStatementExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        request = {
            "Statements": [statement_.api_request for statement_ in self._statements]
        }

        response = self._dispatch_api_call(
            self.connection.client.batch_execute_statement, request
        )
        self.process_rows(response)
        self.post_execute()

    def process_rows(self, response: Dict[str, Any]) -> None:
        rows = response.get("Responses", None)
        if rows is None:
            raise DataError("KeyError `Responses` in BatchStatementResponse")

        processed_error_rows = list()
        processed_rows = list()
        for row in rows:
            error = row.get("Error", None)
            if error:
                processed_error_rows.append(
                    {
                        error.get("Code"): error.get("Message"),
                    }
                )
                continue

            item = row.get("Item", None)
            if item:
                row_ = self._process_undef_row_item(item)
                processed_rows.append(row_)

        self._errors.extend(processed_error_rows)
        self._rows.extend(processed_rows)
        self._next_token = response.get("NextToken", None)


class DmlTransactionExecutor(DmlBatchExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        request = {
            "TransactStatements": [
                statement_.api_request for statement_ in self._statements
            ]
        }

        if self.next_token:
            request.update({"ClientRequestToken": self.next_token})

        response = self._dispatch_api_call(
            self.connection.client.execute_transaction, request
        )
        self.process_rows(response)
        self.post_execute()


class DdlExecutor(BaseExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def process_rows(self, response: Dict[str, Any]) -> None:
        self.rows.extend(
            (k, json.dumps(v, default=self._handle_conversion))
            for k, v in response.items()
        )
        self._metadata.update(ColumnInfo("response_name", "response_name"))
        self._metadata.update(ColumnInfo("response_value", "response_value"))

    def _handle_conversion(self, val: Any) -> str:
        if isinstance(val, datetime):
            return str(val)
        else:
            return val


class DdlCreateExecutor(DdlExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        response = self._dispatch_api_call(self.connection.client.create_table, request)
        self.process_rows(response)
        self.post_execute()


class DdlAlterExecutor(DdlExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        response = self._dispatch_api_call(self.connection.client.update_table, request)
        self.process_rows(response)
        self.post_execute()


class DdlDropExecutor(DdlExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        response = self._dispatch_api_call(self.connection.client.delete_table, request)
        self.process_rows(response)
        self.post_execute()


class DdlCreateGlobalExecutor(DdlExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        response = self._dispatch_api_call(
            self.connection.client.create_global_table, request
        )
        self.process_rows(response)
        self.post_execute()


class DdlDropGlobalExecutor(DdlExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        response = self._dispatch_api_call(
            self.connection.client.update_global_table, request
        )
        self.process_rows(response)
        self.post_execute()


class UtilListTablesExecutor(BaseExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        if self.next_token:
            request.update({"ExclusiveStartTableName": self.next_token})

        response = self._dispatch_api_call(self.connection.client.list_tables, request)
        self.process_rows(response)
        self.post_execute()

    def process_rows(self, response: Dict[str, Any]) -> None:
        tables = response.get("TableNames", None)
        if tables is None:
            raise DataError("KeyError `TableNames`")

        self._rows.extend((t) for t in tables)
        self._next_token = response.get("LastEvaluatedTableName", None)
        self._metadata.update(ColumnInfo("table_name", "table_name"))


class UtilDescTableExecutor(DdlExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        response = self._dispatch_api_call(
            self.connection.client.describe_table, request
        )
        self.process_rows(response)
        self.post_execute()


class UtilListGlobalTablesExecutor(BaseExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        if self.next_token:
            request.update({"ExclusiveStartGlobalTableName": self.next_token})

        response = self._dispatch_api_call(
            self.connection.client.list_global_tables, request
        )
        self.process_rows(response)
        self.post_execute()

    def process_rows(self, response: Dict[str, Any]) -> None:
        tables = response.get("GlobalTables", None)
        if tables is None:
            raise DataError("KeyError `GlobalTables`")

        for table in tables:
            table_name = table["GlobalTableName"]
            regions = ",".join([r["RegionName"] for r in table["ReplicationGroup"]])
            self._rows.append((table_name, regions))

        self._next_token = response.get("LastEvaluatedGlobalTableName", None)
        self._metadata.update(ColumnInfo("table_name", "table_name"))
        self._metadata.update(ColumnInfo("region_names", "region_names"))


class UtilDescGlobalTableExecutor(DdlExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs,
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs,
        )

    def execute(self) -> None:
        statement_ = self._statements[0]
        request = statement_.api_request

        response = self._dispatch_api_call(
            self.connection.client.describe_global_table, request
        )
        self.process_rows(response)



# -*- coding: utf-8 -*-
"""
Syntax of DDB table dropping with SQL:
DROP [GLOBAL] TABLE tbl_name
    [ReplicationGroup [=] ('string', ...)]

Sample SQL of Dropping Table:
-----------------------------
DROP TABLE Issues

DROP GLOBAL TABLE Issues
    ReplicationGroup (us-east-1, us-west-2)
"""
import logging
from .ddl_sql import DdlBase
from .common import KeyWords, Tokens
from pyparsing import Opt, Group, Forward, delimited_list
from typing import Any, Dict

_logger = logging.getLogger(__name__)  # type: ignore


class DdlDrop(DdlBase):
    _DROP_TABLE_STATEMENT = (
        KeyWords.DROP
        + Opt(KeyWords.GLOBAL)("global").set_name("global")
        + KeyWords.TABLE
        + Tokens.TABLE_NAME
        + Opt(
            DdlBase._REPLICATION_GROUP
            + Opt(KeyWords.EQUALS)
            + KeyWords.LPAR
            + Group(delimited_list(Tokens.REGION_NAME))("replication_group").set_name(
                "replication_group"
            )
            + KeyWords.RPAR
        )
    )("drop_statement").set_name("drop_statement")

    _DDL_DROP_EXPR = Forward()
    _DDL_DROP_EXPR <<= _DROP_TABLE_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    @property
    def syntax_def(self) -> Forward:
        return DdlDrop._DDL_DROP_EXPR

    def transform(self) -> Dict[str, Any]:
        if self.root_parse_results is None:
            raise ValueError("Statement was not parsed yet")

        request = dict()
        is_global_table = (
            True if self.root_parse_results.get("global", None) is not None else False
        )
        table_name_ = self.root_parse_results["table"]

        if is_global_table:
            request.update({"GlobalTableName": table_name_})
            replication_group = self.root_parse_results["replication_group"]

            request.update(
                {
                    "ReplicaUpdates": [
                        {"Delete": {"RegionName": r}} for r in replication_group
                    ]
                }
            )
        else:
            request.update({"TableName": table_name_})

        return request



# -*- coding: utf-8 -*-
from ..sql.dml_select import DmlSelect
from ..sql.common import KeyWords
from pyparsing import (
    Forward,
    SkipTo,
)
from typing import Any, Dict


class SupersetSelect(DmlSelect):
    _BASE_SELECT_STATEMENT = DmlSelect._SELECT_STATEMENT

    _INNER_SELECT_STATEMENT = (
        KeyWords.SELECT
        + SkipTo(KeyWords.FROM)("inner_columns").set_name("inner_columns")
        + KeyWords.FROM
        + KeyWords.LPAR
        + _BASE_SELECT_STATEMENT
        + KeyWords.RPAR
        + SkipTo(KeyWords.RPAR)("inner_exprs").set_name("inner_exprs")
    )("inner_select").set_name("inner_select")

    _OUTER_SELECT_STATEMENT = (
        KeyWords.SELECT
        + SkipTo(KeyWords.FROM)("outer_columns").set_name("outer_columns")
        + KeyWords.FROM
        + KeyWords.LPAR
        + (_INNER_SELECT_STATEMENT | _BASE_SELECT_STATEMENT)
        + KeyWords.RPAR
        + SkipTo(KeyWords.SEMICOLON)("outer_exprs").set_name("outer_exprs")
    )("outer_select").set_name("outer_select")

    _NESTED_SELECT_STATEMENT = (
        _OUTER_SELECT_STATEMENT | _INNER_SELECT_STATEMENT | _BASE_SELECT_STATEMENT
    )("nested_select_statement").set_name("nested_select_statement")

    _SUPERSET_SELECT_EXPR = Forward()
    _SUPERSET_SELECT_EXPR <<= _NESTED_SELECT_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)
        self._outer_columns = None
        self._outer_exprs = None
        self._inner_columns = None
        self._inner_exprs = None
        self._is_nested = False

    def preprocess(self) -> None:
        self._executed_statement = self._executed_statement.strip()
        if self._executed_statement[-1] != ";":
            self._executed_statement += ";"

    @property
    def outer_columns(self) -> str:
        return self._outer_columns

    @property
    def outer_exprs(self) -> str:
        return self._outer_exprs

    @property
    def inner_columns(self) -> str:
        return self._inner_columns

    @property
    def inner_exprs(self) -> str:
        return self._inner_exprs

    @property
    def is_nested(self) -> bool:
        return self._is_nested

    @property
    def syntax_def(self) -> Forward:
        return SupersetSelect._SUPERSET_SELECT_EXPR

    def transform(self) -> Dict[str, Any]:
        converted_ = super().transform()

        outer_columns = self.root_parse_results.get("outer_columns", None)
        if outer_columns is not None:
            self._outer_columns = outer_columns.strip()
            self._is_nested = True

        outer_exprs = self.root_parse_results.get("outer_exprs", None)
        if outer_exprs is not None:
            self._outer_exprs = outer_exprs.strip()

        inner_columns = self.root_parse_results.get("inner_columns", None)
        if inner_columns is not None:
            self._inner_columns = inner_columns.strip()

        inner_exprs = self.root_parse_results.get("inner_exprs", None)
        if inner_exprs is not None:
            self._inner_exprs = inner_exprs.strip()

        return converted_



# -*- coding: utf-8 -*-
import logging
import ast
from .sql.common import Functions
from datetime import datetime, date
from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Dict, Optional, Set, List, Union, Type

_logger = logging.getLogger(__name__)  # type: ignore


class Serializer(metaclass=ABCMeta):
    def __init__(
        self,
        mapping: Optional[Dict[Any, Callable[[Optional[str]], Optional[Any]]]] = None,
    ) -> None:
        if mapping is None:
            self._mapping = self.get_default_converters()
        else:
            self._mapping = mapping

    @property
    def mapping(self) -> Dict[str, Callable[[Optional[str]], Optional[Any]]]:
        return self._mapping

    def _to_string(self, value: Optional[str], **kwargs) -> Optional[Dict[str, str]]:
        if value is None:
            return None
        return {"S": value}

    def _to_number(
        self, value: Optional[Union[int, float]], **kwargs
    ) -> Optional[Dict[str, str]]:
        if value is None:
            return None
        return {"N": str(value)}

    def _to_binary(self, value: Optional[bytes], **kwargs) -> Optional[Dict[str, str]]:
        if value is None:
            return None
        return {"B": value.decode()}

    def _to_set(
        self, value: Optional[Set[Any]], **kwargs
    ) -> Optional[Dict[str, List[str]]]:
        if value is None:
            return None

        value_ = next(iter(value))
        if isinstance(value_, type(1)) or isinstance(value_, type(1.0)):
            return {"NS": [str(v) for v in value]}
        elif isinstance(value_, type(b"")):
            return {"BS": [v.decode() for v in value]}
        else:
            return {"SS": [v for v in value]}

    def _to_map(
        self, value: Optional[Dict[str, Any]], **kwargs
    ) -> Optional[Dict[str, Any]]:
        if value is None:
            return None

        converted_ = {}
        for k, v in value.items():
            type_ = type(v)
            converted_[k] = self._mapping.get(type_, None)(v)
        return {"M": converted_}

    def _to_list(
        self, value: Optional[List[Any]], **kwargs
    ) -> Optional[Dict[str, Any]]:
        if value is None:
            return None

        converted_ = []
        for v in value:
            type_ = type(v)
            converted_.append(self._mapping.get(type_, None)(v))
        return {"L": converted_}

    def _to_null(self, value: Optional[Any], **kwargs) -> Optional[Dict[str, Any]]:
        return {"NULL": False if value else True}

    def _to_bool(self, value: Optional[bool], **kwargs) -> Optional[Dict[str, Any]]:
        return {"BOOL": value}

    def _to_datetime(
        self, value: Optional[Union[datetime, date]], **kwargs
    ) -> Optional[Dict[str, str]]:
        if value is None:
            return None

        return {"S": value.isoformat()}

    def _to_default(self, value: Optional[Any], **kwargs) -> Optional[Dict[str, Any]]:
        return {"S": str(value)}

    def get(self, type_: Type) -> Callable[[Optional[str]], Optional[Any]]:
        return self._mapping.get(type_, self._to_default)

    def get_default_converters(
        self,
    ) -> Dict[Type[Any], Callable[[Optional[str]], Optional[Any]]]:
        return {
            str: self._to_string,
            int: self._to_number,
            float: self._to_number,
            bytes: self._to_binary,
            set: self._to_set,
            dict: self._to_map,
            list: self._to_list,
            type(None): self._to_null,
            bool: self._to_bool,
            datetime: self._to_datetime,
            date: self._to_datetime,
        }


class Deserializer(metaclass=ABCMeta):
    def __init__(
        self,
        mapping: Optional[Dict[Any, Callable[[Optional[str]], Optional[Any]]]] = None,
    ) -> None:
        if mapping is None:
            self._mapping = self.get_default_converters()
        else:
            self._mapping = mapping

    @property
    def mapping(self) -> Dict[str, Callable[[Optional[str]], Optional[Any]]]:
        return self._mapping

    def _to_string(self, value: Optional[Any], **kwargs) -> Optional[Any]:
        function_ = kwargs.get("function", None)

        if function_ == Functions.DATE:
            return self._to_date(value, **kwargs)
        elif function_ == Functions.DATETIME:
            return self._to_datetime(value, **kwargs)
        elif function_ == Functions.SUBSTR or function_ == Functions.SUBSTRING:
            return self._to_substr(value, **kwargs)
        elif function_ == Functions.TRIM:
            return value.strip()
        elif function_ == Functions.LTRIM:
            return value.lstrip()
        elif function_ == Functions.RTRIM:
            return value.rstrip()
        elif function_ == Functions.UPPER:
            return value.upper()
        elif function_ == Functions.LOWER:
            return value.lower()
        elif function_ == Functions.REPLACE:
            return self._to_replace(value, **kwargs)
        else:
            return value

    def _to_date(self, value: Optional[Any], **kwargs) -> Optional[datetime]:
        function_params_ = kwargs.get("function_params", None)
        if function_params_ is None or len(function_params_) == 0:
            return date.fromisoformat(value)
        else:
            return datetime.strptime(value, function_params_[0]).date()

    def _to_datetime(self, value: Optional[Any], **kwargs) -> Optional[datetime]:
        function_params_ = kwargs.get("function_params", None)
        if function_params_ is None or len(function_params_) == 0:
            return datetime.fromisoformat(value)
        else:
            return datetime.strptime(value, function_params_[0])

    def _to_number(self, value: Optional[str], **kwargs) -> Optional[Union[int, float]]:
        if value is None:
            return None
        return ast.literal_eval(value)

    def _to_binary(self, value: Optional[str], **kwargs) -> Optional[bytes]:
        if value is None:
            return None
        return value

    def _to_string_set(
        self, value: Optional[List[str]], **kwargs
    ) -> Optional[Set[str]]:
        if value is None:
            return None
        return set([v for v in value])

    def _to_number_set(
        self, value: Optional[List[str]], **kwargs
    ) -> Optional[Set[float]]:
        if value is None:
            return None
        return set([float(v) for v in value])

    def _to_binary_set(
        self, value: Optional[List[str]], **kwargs
    ) -> Optional[Set[bytes]]:
        return self._to_string_set(value)

    def _to_map(
        self, value: Optional[Dict[str, Any]], **kwargs
    ) -> Optional[Dict[str, Any]]:
        if value is None:
            return None

        converted_ = {}
        for k, v in value.items():
            type_, value_ = next(iter(v.items()))
            converted_[k] = self._mapping.get(type_, None)(value_)
        return converted_

    def _to_list(self, value: Optional[List[Any]], **kwargs) -> Optional[List[Any]]:
        if value is None:
            return None

        coverted_ = []
        for v in value:
            type_, value_ = next(iter(v.items()))
            coverted_.append(self._mapping.get(type_, None)(value_))
        return coverted_

    def _to_null(self, value: Optional[bool], **kwargs) -> Optional[bool]:
        return None

    def _to_bool(self, value: Optional[bool], **kwargs) -> Optional[bool]:
        return value

    def _to_substr(self, value: Optional[str], **kwargs) -> Optional[str]:
        function_params_ = kwargs.get("function_params", None)
        if function_params_ is None or len(function_params_) == 0:
            return value
        elif len(function_params_) == 1:
            start = int(function_params_[0])
            return value[start:]
        else:
            start = int(function_params_[0])
            length = int(function_params_[1])
            end = start + length if start + length < len(value) else len(value)
            return value[start:end]

    def _to_replace(self, value: Optional[str], **kwargs) -> Optional[str]:
        function_params_ = kwargs.get("function_params", None)
        if function_params_ is None or len(function_params_) == 0:
            return value
        elif len(function_params_) == 1:
            return value.replace(function_params_[0], "")
        else:
            return value.replace(function_params_[0], function_params_[1])

    def _to_default(self, value: Optional[Any], **kwargs) -> Optional[str]:
        return value

    def get(self, type_: str) -> Callable[[Optional[str]], Optional[Any]]:
        return self._mapping.get(type_, self._to_default)

    def get_default_converters(
        self,
    ) -> Dict[Any, Callable[[Optional[str]], Optional[Any]]]:
        return {
            "S": self._to_string,
            "N": self._to_number,
            "B": self._to_binary,
            "SS": self._to_string_set,
            "NS": self._to_number_set,
            "BS": self._to_binary_set,
            "M": self._to_map,
            "L": self._to_list,
            "NULL": self._to_null,
            "BOOL": self._to_bool,
        }


class Converter(metaclass=ABCMeta):
    def __init__(
        self,
        serializer: Serializer,
        deserializer: Deserializer,
    ) -> None:
        self._serializer = serializer
        self._deserializer = deserializer

    @property
    def serializer(self) -> Serializer:
        return self._serializer

    @property
    def deserializer(self) -> Deserializer:
        return self._deserializer

    def get_serialize_converter(
        self, type_: str
    ) -> Callable[[Optional[str]], Optional[Any]]:
        return self._serializer.get(type_)

    def get_deserialize_converter(
        self, type_: str
    ) -> Callable[[Optional[str]], Optional[Any]]:
        return self._deserializer.get(type_)

    @abstractmethod
    def serialize(self, value: Optional[Any], **kwargs) -> Optional[Any]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def deserialize(self, value: Optional[Any], **kwargs) -> Optional[Any]:
        raise NotImplementedError  # pragma: no cover


class DefaultTypeConverter(Converter):
    def __init__(self) -> None:
        super().__init__(
            serializer=Serializer(),
            deserializer=Deserializer(),
        )

    def serialize(self, value: Optional[Any], **kwargs) -> Optional[Any]:
        type_ = type(value)
        converter = self.get_serialize_converter(type_)
        return converter(value, **kwargs)

    def deserialize(self, value: Optional[Any], **kwargs) -> Optional[Any]:
        type_, value_ = next(iter(value.items()))
        converter = self.get_deserialize_converter(type_)
        return converter(value_, **kwargs)



# -*- coding: utf-8 -*-
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, cast

from .converter import Converter
from .common import CursorIterator
from .model import Statements, Metadata
from .executor import BaseExecutor, dispatch_executor
from .executor import DmlStatementDictExecutor
from .error import ProgrammingError
from .util import RetryConfig

if TYPE_CHECKING:
    from .connection import Connection

_logger = logging.getLogger(__name__)  # type: ignore


class DynamoDBResultSet(CursorIterator):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        arraysize: int,
        retry_config: RetryConfig,
        is_transaction: bool = False,
        executor_class: BaseExecutor = None,
        **kwargs
    ) -> None:
        super().__init__(arraysize=arraysize)
        assert statements and len(statements) > 0, "Required statements not found."
        self._connection: Optional["Connection"] = connection
        self._arraysize = arraysize
        self._statements = statements
        self._executor = dispatch_executor(
            connection,
            converter,
            statements,
            retry_config,
            is_transaction,
            executor_class=executor_class,
            **kwargs
        )
        assert self._executor is not None, "Executor is not specified"

    @property
    def connection(self) -> "Connection":
        if self.is_closed:
            raise ProgrammingError("DynamoDBResultSet is closed.")
        return cast("Connection", self._connection)

    @property
    def errors(self) -> List[Dict[str, str]]:
        return self._executor.errors

    @property
    def metadata(self) -> Optional[Metadata]:
        return self._executor.metadata

    @property
    def rowcount(self) -> int:
        return len(self._executor.rows)

    @property
    def description(
        self,
    ) -> Optional[List[Tuple[str, str, None, None, None, None, None]]]:
        if self._executor.metadata is None:
            return None
        return [
            (
                (
                    column_info.alias
                    if column_info.alias is not None
                    else column_info.name
                ),
                column_info.type_code,
                None,
                None,
                None,
                None,
                None,
            )
            for column_info in self._executor.metadata
        ]

    def fetchone(
        self,
    ) -> Optional[Dict[Any, Optional[Any]]]:
        limit_ = self._statements.limit
        while len(self._executor.rows) == 0 and self._executor.next_token:
            if not limit_ or (limit_ and self._rownumber < limit_):
                self._executor.execute()
            else:
                break
        if not self._executor.rows:
            return None
        else:
            if self._rownumber is None:
                self._rownumber = 0
            if limit_ and self._rownumber >= limit_:
                self._executor.rows.clear()
                return None
            row = self._executor.rows.popleft()
            self._rownumber += 1
            return row

    def fetchmany(self, size: Optional[int] = None) -> List[Dict[Any, Optional[Any]]]:
        if not size or size <= 0:
            size = self._arraysize
        rows = []
        for _ in range(size):
            row = self.fetchone()
            if row:
                rows.append(row)
            else:
                break
        return rows

    def fetchall(self) -> List[Dict[Any, Optional[Any]]]:
        rows = []
        while True:
            row = self.fetchone()
            if row:
                rows.append(row)
            else:
                break
        return rows

    @property
    def is_closed(self) -> bool:
        return self._connection is None

    def close(self) -> None:
        self._connection = None
        self._executor = None
        self._rownumber = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class DynamoDBDictResultSet(DynamoDBResultSet):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        arraysize: int,
        retry_config: RetryConfig,
        is_transaction: bool = False,
        executor_class: BaseExecutor = None,
        **kwargs
    ) -> None:
        assert (
            statements.query_type[0] == "DML" and len(statements) == 1
        ), "DictExecutor can only support single DML statement"
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            arraysize=arraysize,
            retry_config=retry_config,
            is_transaction=is_transaction,
            executor_class=DmlStatementDictExecutor,
            **kwargs
        )



# -*- coding: utf-8 -*-
import logging
import time
from boto3.session import Session
from botocore.config import Config
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

from .converter import Converter
from .cursor import BaseCursor, Cursor
from .error import NotSupportedError
from .util import RetryConfig, retry_api_call


if TYPE_CHECKING:
    from botocore.client import BaseClient

_logger = logging.getLogger(__name__)


class Connection:
    _SESSION_PASSING_ARGS: List[str] = [
        "aws_access_key_id",
        "aws_secret_access_key",
        "aws_session_token",
        "region_name",
        "botocore_session",
        "profile_name",
    ]
    _CLIENT_PASSING_ARGS: List[str] = [
        "aws_access_key_id",
        "aws_secret_access_key",
        "aws_session_token",
        "config",
        "api_version",
        "use_ssl",
        "verify",
        "endpoint_url",
    ]

    def __init__(
        self,
        region_name: Optional[str] = None,
        profile_name: Optional[str] = None,
        role_arn: Optional[str] = None,
        role_session_name: str = f"PyDynamoDB-session-{int(time.time())}",
        external_id: Optional[str] = None,
        serial_number: Optional[str] = None,
        token_code: Optional[str] = None,
        principal_arn: Optional[str] = None,
        saml_assertion: Optional[str] = None,
        web_identity_token: Optional[str] = None,
        provider_id: Optional[str] = None,
        duration_seconds: int = 3600,
        session: Optional[Session] = None,
        config: Optional[Config] = None,
        converter: Optional[Converter] = None,
        retry_config: Optional[RetryConfig] = None,
        cursor_class: Type[BaseCursor] = Cursor,
        cursor_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        self._kwargs = {
            **kwargs,
            "role_arn": role_arn,
            "role_session_name": role_session_name,
            "external_id": external_id,
            "serial_number": serial_number,
            "token_code": token_code,
            "principal_arn": principal_arn,
            "saml_assertion": saml_assertion,
            "web_identity_token": web_identity_token,
            "provider_id": provider_id,
            "duration_seconds": duration_seconds,
        }

        self.region_name = region_name
        self.profile_name = profile_name
        self.config: Optional[Config] = config if config else Config()

        if session:
            self._session = session
        else:
            creds = None
            if role_arn and saml_assertion:
                creds = self._assume_role_with_saml(
                    profile_name=self.profile_name,
                    region_name=self.region_name,
                    role_arn=role_arn,
                    principal_arn=principal_arn,
                    saml_assertion=saml_assertion,
                    duration_seconds=duration_seconds,
                )
            elif role_arn and web_identity_token:
                creds = self._assume_role_with_web_identity(
                    profile_name=self.profile_name,
                    region_name=self.region_name,
                    role_arn=role_arn,
                    role_session_name=role_session_name,
                    web_identity_token=web_identity_token,
                    provider_id=provider_id,
                    duration_seconds=duration_seconds,
                )
            elif role_arn:
                creds = self._assume_role(
                    profile_name=self.profile_name,
                    region_name=self.region_name,
                    role_arn=role_arn,
                    role_session_name=role_session_name,
                    external_id=external_id,
                    serial_number=serial_number,
                    token_code=token_code,
                    duration_seconds=duration_seconds,
                )
            elif serial_number:
                creds = self._get_session_token(
                    profile_name=self.profile_name,
                    region_name=self.region_name,
                    serial_number=serial_number,
                    duration_seconds=duration_seconds,
                )

            if creds:
                self.profile_name = None
                self._kwargs.update(
                    {
                        "aws_access_key_id": creds["AccessKeyId"],
                        "aws_secret_access_key": creds["SecretAccessKey"],
                        "aws_session_token": creds["SessionToken"],
                    }
                )
            self._session = Session(
                region_name=self.region_name,
                profile_name=self.profile_name,
                **self._session_kwargs,
            )

        self._client = self._session.client(
            "dynamodb",
            region_name=self.region_name,
            config=self.config,
            **self._client_kwargs,
        )

        self._converter = converter
        self._retry_config = retry_config if retry_config else RetryConfig()
        self.cursor_class = cursor_class
        self.cursor_kwargs = cursor_kwargs if cursor_kwargs else dict()
        self._cursor_pool = list()
        self._autocommit = True
        self._in_transaction = False

    def _assume_role(
        self,
        profile_name: Optional[str],
        region_name: Optional[str],
        role_arn: str,
        role_session_name: str,
        external_id: Optional[str],
        serial_number: Optional[str],
        token_code: Optional[str],
        duration_seconds: int,
    ) -> Dict[str, Any]:
        session = Session(
            region_name=region_name, profile_name=profile_name, **self._session_kwargs
        )
        client = session.client(
            "sts", region_name=region_name, config=self.config, **self._client_kwargs
        )
        request = {
            "RoleArn": role_arn,
            "RoleSessionName": role_session_name,
            "DurationSeconds": duration_seconds,
        }
        if external_id:
            request.update(
                {
                    "ExternalId": external_id,
                }
            )
        if serial_number:
            request.update(
                {
                    "SerialNumber": serial_number,
                    "TokenCode": token_code,
                }
            )

        response = client.assume_role(**request)
        creds: Dict[str, Any] = response["Credentials"]
        return creds

    def _assume_role_with_saml(
        self,
        profile_name: Optional[str],
        region_name: Optional[str],
        role_arn: str,
        principal_arn: str,
        saml_assertion: str,
        duration_seconds: int,
    ) -> Dict[str, Any]:
        session = Session(
            region_name=region_name, profile_name=profile_name, **self._session_kwargs
        )
        client = session.client(
            "sts", region_name=region_name, config=self.config, **self._client_kwargs
        )

        # Resolve the issue which plus sign will be replaced to space by url.parse_qsl
        saml_assertion = saml_assertion.replace(" ", "+")

        request = {
            "RoleArn": role_arn,
            "PrincipalArn": principal_arn,
            "SAMLAssertion": saml_assertion,
            "DurationSeconds": duration_seconds,
        }

        response = client.assume_role_with_saml(**request)
        creds: Dict[str, Any] = response["Credentials"]
        return creds

    def _assume_role_with_web_identity(
        self,
        profile_name: Optional[str],
        region_name: Optional[str],
        role_arn: str,
        role_session_name: str,
        web_identity_token: str,
        provider_id: Optional[str],
        duration_seconds: int,
    ) -> Dict[str, Any]:
        session = Session(
            region_name=region_name, profile_name=profile_name, **self._session_kwargs
        )
        client = session.client(
            "sts", region_name=region_name, config=self.config, **self._client_kwargs
        )
        request = {
            "RoleArn": role_arn,
            "RoleSessionName": role_session_name,
            "DurationSeconds": duration_seconds,
            "WebIdentityToken": web_identity_token,
        }

        if provider_id:
            request.update(
                {
                    "ProviderId": provider_id,
                }
            )

        response = client.assume_role_with_web_identity(**request)
        creds: Dict[str, Any] = response["Credentials"]
        return creds

    def _get_session_token(
        self,
        profile_name: Optional[str],
        region_name: Optional[str],
        serial_number: Optional[str],
        duration_seconds: int,
    ) -> Dict[str, Any]:
        session = Session(profile_name=profile_name, **self._session_kwargs)
        client = session.client(
            "sts", region_name=region_name, config=self.config, **self._client_kwargs
        )
        token_code = input("Enter the MFA code: ")
        request = {
            "DurationSeconds": duration_seconds,
            "SerialNumber": serial_number,
            "TokenCode": token_code,
        }
        response = client.get_session_token(**request)
        creds: Dict[str, Any] = response["Credentials"]
        return creds

    @property
    def _session_kwargs(self) -> Dict[str, Any]:
        return {
            k: v for k, v in self._kwargs.items() if k in self._SESSION_PASSING_ARGS
        }

    @property
    def _client_kwargs(self) -> Dict[str, Any]:
        return {k: v for k, v in self._kwargs.items() if k in self._CLIENT_PASSING_ARGS}

    @property
    def session(self) -> Session:
        return self._session

    @property
    def client(self) -> "BaseClient":
        return self._client

    @property
    def autocommit(self) -> bool:
        return self._autocommit

    @autocommit.setter
    def autocommit(self, value: bool) -> None:
        try:
            if not self._autocommit and value:
                self._autocommit = True
                for cursor_ in self.cursor_pool:
                    cursor_.flush()
        finally:
            self._autocommit = value

    @property
    def in_transaction(self) -> bool:
        return self._in_transaction

    @in_transaction.setter
    def in_transaction(self, value: bool) -> bool:
        self._in_transaction = False

    @property
    def cursor_pool(self) -> List[Optional[BaseCursor]]:
        return self._cursor_pool

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def cursor(self, cursor: Optional[Type[BaseCursor]] = None, **kwargs) -> BaseCursor:
        kwargs.update(self.cursor_kwargs)
        if not cursor:
            cursor = self.cursor_class
        converter = kwargs.pop("converter", self._converter)
        if not converter:
            converter = cursor.get_default_converter()
        cursor_ = cursor(
            connection=self,
            converter=converter,
            retry_config=kwargs.pop("retry_config", self._retry_config),
            **kwargs,
        )

        self._cursor_pool.append(cursor_)
        return cursor_

    def close(self) -> None:
        self._session = None
        self._client = None
        self._autocommit = True
        self._in_transaction = False
        self.cursor_pool.clear()

    def begin(self) -> None:
        self._autocommit = False
        self._in_transaction = True

    def commit(self) -> None:
        try:
            if self._in_transaction:
                for cursor_ in self.cursor_pool:
                    cursor_.execute_transaction()
        finally:
            self._autocommit = True
            self._in_transaction = False

    def rollback(self) -> None:
        raise NotSupportedError

    def test_connection(self) -> bool:
        request: Dict[str, Any] = {"Limit": 1}

        try:
            retry_api_call(
                self._client.list_tables,
                config=self._retry_config,
                logger=_logger,
                **request,
            )
        except Exception as e:
            _logger.exception("Failed to connect database: %s" % repr(e))
            return False
        else:
            return True



# -*- coding: utf-8 -*-
import re
from ..util import strtobool
from pyparsing import (
    Word,
    nums,
    alphanums,
    Combine,
    Regex,
    CaselessKeyword,
    Opt,
    one_of,
    Group,
    Suppress,
    pyparsing_common as ppc,
)


class DataTypes:
    STRING = "STRING"
    DATE = "DATE"
    DATETIME = "DATETIME"
    NUMBER = "NUMBER"
    BOOL = "BOOL"
    NULL = "NULL"


class DdbDataTypes:
    BOOLEAN = "BOOL"
    BINARY = "B"
    LIST = "L"
    MAP = "M"
    NUMBER = "N"
    NUMBER_SET = "NS"
    STRING_SET = "SS"
    STRING = "STRING"
    NULL = "NULL"


class Functions:
    TYPE_CONVERSION = "TYPE_CONVERSION"
    STRING_FUNCTION = "STRING_FUNCTION"

    DATE = "DATE"
    DATETIME = "DATETIME"
    NUMBER = "NUMBER"
    BOOL = "BOOL"
    SUBSTR = "SUBSTR"
    SUBSTRING = "SUBSTRING"
    TRIM = "TRIM"
    LTRIM = "LTRIM"
    RTRIM = "RTRIM"
    REPLACE = "REPLACE"
    UPPER = "UPPER"
    LOWER = "LOWER"

    SUPPORTED_FUNTIONS = {
        TYPE_CONVERSION: [
            DATE,
            DATETIME,
            NUMBER,
            BOOL,
        ],
        STRING_FUNCTION: [
            SUBSTR,
            SUBSTRING,
            TRIM,
            RTRIM,
            LTRIM,
            REPLACE,
            UPPER,
            LOWER,
        ],
    }
    WHERE_CONDITION = ["begins_with", "attribute_type", "contains"]
    WHERE_CONDITION_WITH_OP = ["size"]


class KeyWords:
    # Common sign definition
    (
        LPAR,
        RPAR,
        COMMA,
        SEMICOLON,
        COLON,
        DOT,
        SINGLEQUOTE,
        DOUBLEQUOTE,
        BACKQUOTE,
        EQUALS,
        SPACE,
        LBRACKET,
        RBRACKET,
        LCURLYBRACKET,
        RCURLYBRACKET,
    ) = map(Suppress, "(),;:.'\"`= []{}")
    STAR, QUESTION = "*", "?"
    SUPPRESS_QUOTE = (SINGLEQUOTE | BACKQUOTE | DOUBLEQUOTE).set_name("suppress_quote")
    QUOTE = one_of("' \" `").set_name("quote")

    # Operators
    (
        ADD,
        SUBTRACT,
        EQUAL_TO,
        NOT_EQUAL_TO,
        GREATER,
        LESS,
        GREATER_OR_EQUAL,
        LESS_OR_EQUAL,
        AND,
        BETWEEN,
        IN,
        IS,
        NOT,
        OR,
    ) = map(
        CaselessKeyword,
        [
            "+",
            "-",
            "=",
            "<>",
            ">",
            "<",
            ">=",
            "<=",
            "AND",
            "BETWEEN",
            "IN",
            "IS",
            "NOT",
            "OR",
        ],
    )
    ARITHMETIC_OPERATORS = one_of("+ -")("arithmetic_operators").set_name(
        "arithmetic_operators"
    )
    COMPARISON_OPERATORS = one_of("= <> != < > >= <=")("comparison_operators").set_name(
        "comparison_operators"
    )

    # Common key word definition
    (
        CREATE,
        ALTER,
        DROP,
        TABLE,
        SELECT,
        FROM,
        WHERE,
        INSERT,
        INTO,
        VALUE,
        GLOBAL,
        PARTITION_KEY,
        SORT_KEY,
        HASH,
        RANGE,
        UPDATE,
        SET,
        REMOVE,
        DELETE,
        INDEX,
        REPLICA,
        ORDER_BY,
        LIMIT,
        AS,
        RETURNING,
    ) = map(
        CaselessKeyword,
        [
            "CREATE",
            "ALTER",
            "DROP",
            "TABLE",
            "SELECT",
            "FROM",
            "WHERE",
            "INSERT",
            "INTO",
            "VALUE",
            "GLOBAL",
            "PARTITION KEY",
            "SORT KEY",
            "HASH",
            "RANGE",
            "UPDATE",
            "SET",
            "REMOVE",
            "DELETE",
            "INDEX",
            "REPLICA",
            "ORDER BY",
            "Limit",
            "AS",
            "RETURNING",
        ],
    )

    (
        LIST,
        SHOW,
        DESC,
        DESCRIBE,
        TABLES,
    ) = map(
        CaselessKeyword,
        [
            "LIST",
            "SHOW",
            "DESC",
            "DESCRIBE",
            "TABLES",
        ],
    )

    # Functions supported on Column
    FUNCTION_ON_COLUMN = one_of(
        [fun for subfuns in Functions.SUPPORTED_FUNTIONS.values() for fun in subfuns],
        caseless=True,
        as_keyword=True,
    )("function").set_name("function")

    # Functions supported on Where conditions
    FUNCTION_ON_WHERE = one_of(
        Functions.WHERE_CONDITION, caseless=True, as_keyword=True
    )("function").set_name("function")

    FUNCTION_WITH_OP_ON_WHERE = one_of(
        Functions.WHERE_CONDITION_WITH_OP, caseless=True, as_keyword=True
    )("function_with_op").set_name("function_with_op")


class Tokens:
    # Common token definition
    TABLE_NAME = (
        Opt(KeyWords.SUPPRESS_QUOTE)
        + Word(alphanums + "-_.")("table").set_name("table")
        + Opt(KeyWords.SUPPRESS_QUOTE)
    )
    ATTRIBUTE_NAME = (
        Opt(KeyWords.SUPPRESS_QUOTE)
        + Word(alphanums + "-_.")("attribute_name").set_name("attribute_name")
        + Opt(KeyWords.SUPPRESS_QUOTE)
    )
    DATA_TYPE = one_of("NUMERIC STRING BINARY", caseless=True)("data_type").set_name(
        "data_type"
    )
    OPT_KEY_TYPE = one_of(["PARTITION KEY", "SORT KEY", "HASH", "RANGE", ""])(
        "key_type"
    ).set_name("key_type")
    KEY_TYPE = one_of(["PARTITION KEY", "SORT KEY", "HASH", "RANGE"])(
        "key_type"
    ).set_name("key_type")
    INDEX_NAME = (
        Opt(KeyWords.SUPPRESS_QUOTE)
        + Word(alphanums + "-_.")("index_name").set_name("index_name")
        + Opt(KeyWords.SUPPRESS_QUOTE)
    )
    INDEX_TYPE = one_of("GLOBAL LOCAL", caseless=True)("index_type").set_name(
        "index_type"
    )
    TAG = Group(
        Opt(KeyWords.SINGLEQUOTE)
        + Word(alphanums + "._$")("tag_key").set_name("tag_key")
        + Opt(KeyWords.SINGLEQUOTE)
        + ":"
        + Opt(KeyWords.SINGLEQUOTE)
        + Word(alphanums + "._$")("tag_value").set_name("tag_value")
        + Opt(KeyWords.SINGLEQUOTE)
    )("tag").set_name("tag")
    REGION_NAME = Word(alphanums + "-")("region_name").set_name("region_name")
    KMSMASTERKEYID = Word(alphanums + "+-=*_$")
    INT_VALUE = Word(nums).set_parse_action(ppc.convert_to_integer)
    BOOL_VALUE = one_of("True False", caseless=True).set_parse_action(
        lambda toks: strtobool(toks[0])
    )
    ORDER_BY_VALUE = one_of("DESC ASC", caseless=True)
    PARTIQL_DATA_TYPE = one_of(
        [
            DdbDataTypes.BOOLEAN,
            DdbDataTypes.BINARY,
            DdbDataTypes.LIST,
            DdbDataTypes.MAP,
            DdbDataTypes.NUMBER,
            DdbDataTypes.NUMBER_SET,
            DdbDataTypes.STRING_SET,
            DdbDataTypes.STRING,
            DdbDataTypes.NULL,
            "MISSING",
        ],
        caseless=True,
    )
    QUOTED_STRING = Combine(
        Suppress('"')
        + Regex(r'(?:[^"\n\r\\]|(?:"")|(?:\\(?:[^x]|x[0-9a-fA-F]+)))*')
        + Suppress('"')
        | Suppress("'")
        + Regex(r"(?:[^'\n\r\\]|(?:'')|(?:\\(?:[^x]|x[0-9a-fA-F]+)))*")
        + Suppress("'")
    )


class QueryType:
    """
    Types of supported queries
    """

    CREATE = ("DDL", "WRITE", "CREATE TABLE")
    CREATE_GLOBAL = ("DDL", "WRITE", "CREATE GLOBAL TABLE")
    ALTER = ("DDL", "WRITE", "ALTER TABLE")
    DROP = ("DDL", "WRITE", "DROP TABLE")
    DROP_GLOBAL = ("DDL", "WRITE", "DROP GLOBAL TABLE")
    INSERT = ("DML", "WRITE", "INSERT")
    UPDATE = ("DML", "WRITE", "UPDATE")
    DELETE = ("DML", "WRITE", "DELETE")
    SELECT = ("DML", "READ", "SELECT")
    LIST = ("UTIL", "READ", "LIST TABLES")
    LIST_GLOBAL = ("UTIL", "READ", "LIST GLOBAL TABLES")
    DESC = ("UTIL", "READ", "DESC")
    DESC_GLOBAL = ("UTIL", "READ", "DESC GLOBAL")


SUPPORTED_QUERY_TYPES = {
    r"^\s*(SELECT).*": QueryType.SELECT,
    r"^\s*(INSERT).*": QueryType.INSERT,
    r"^\s*(UPDATE).*": QueryType.UPDATE,
    r"^\s*(DELETE).*": QueryType.DELETE,
    r"^\s*(CREATE)\s+(TABLE).*": QueryType.CREATE,
    r"^\s*(CREATE)\s+(GLOBAL)\s+(TABLE).*": QueryType.CREATE_GLOBAL,
    r"^\s*(ALTER)\s+(TABLE).*": QueryType.ALTER,
    r"^\s*(DROP)\s+(TABLE).*": QueryType.DROP,
    r"^\s*(DROP)\s+(GLOBAL)\s+(TABLE).*": QueryType.DROP_GLOBAL,
    r"^\s*(LIST|SHOW)\s+(TABLES).*": QueryType.LIST,
    r"^\s*(LIST|SHOW)\s+(GLOBAL)\s+(TABLES).*": QueryType.LIST_GLOBAL,
    r"^\s*(DESC|DESCRIBE)\s+(?!.*GLOBAL)": QueryType.DESC,
    r"^\s*(DESC|DESCRIBE)\s+(GLOBAL).*": QueryType.DESC_GLOBAL,
}


def get_query_type(sql: str) -> QueryType:
    for type_pattern, type in SUPPORTED_QUERY_TYPES.items():
        pattern_ = re.compile(type_pattern, re.IGNORECASE)
        match_ = pattern_.search(sql)

        if match_:
            return type

    raise LookupError("Not supported query type")


# DynamoDB reserved words
# https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/ReservedWords.html
RESERVED_WORDS = [
    "ABORT",
    "ABSOLUTE",
    "ACTION",
    "ADD",
    "AFTER",
    "AGENT",
    "AGGREGATE",
    "ALL",
    "ALLOCATE",
    "ALTER",
    "ANALYZE",
    "AND",
    "ANY",
    "ARCHIVE",
    "ARE",
    "ARRAY",
    "AS",
    "ASC",
    "ASCII",
    "ASENSITIVE",
    "ASSERTION",
    "ASYMMETRIC",
    "AT",
    "ATOMIC",
    "ATTACH",
    "ATTRIBUTE",
    "AUTH",
    "AUTHORIZATION",
    "AUTHORIZE",
    "AUTO",
    "AVG",
    "BACK",
    "BACKUP",
    "BASE",
    "BATCH",
    "BEFORE",
    "BEGIN",
    "BETWEEN",
    "BIGINT",
    "BINARY",
    "BIT",
    "BLOB",
    "BLOCK",
    "BOOLEAN",
    "BOTH",
    "BREADTH",
    "BUCKET",
    "BULK",
    "BY",
    "BYTE",
    "CALL",
    "CALLED",
    "CALLING",
    "CAPACITY",
    "CASCADE",
    "CASCADED",
    "CASE",
    "CAST",
    "CATALOG",
    "CHAR",
    "CHARACTER",
    "CHECK",
    "CLASS",
    "CLOB",
    "CLOSE",
    "CLUSTER",
    "CLUSTERED",
    "CLUSTERING",
    "CLUSTERS",
    "COALESCE",
    "COLLATE",
    "COLLATION",
    "COLLECTION",
    "COLUMN",
    "COLUMNS",
    "COMBINE",
    "COMMENT",
    "COMMIT",
    "COMPACT",
    "COMPILE",
    "COMPRESS",
    "CONDITION",
    "CONFLICT",
    "CONNECT",
    "CONNECTION",
    "CONSISTENCY",
    "CONSISTENT",
    "CONSTRAINT",
    "CONSTRAINTS",
    "CONSTRUCTOR",
    "CONSUMED",
    "CONTINUE",
    "CONVERT",
    "COPY",
    "CORRESPONDING",
    "COUNT",
    "COUNTER",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT",
    "CURSOR",
    "CYCLE",
    "DATA",
    "DATABASE",
    "DATE",
    "DATETIME",
    "DAY",
    "DEALLOCATE",
    "DEC",
    "DECIMAL",
    "DECLARE",
    "DEFAULT",
    "DEFERRABLE",
    "DEFERRED",
    "DEFINE",
    "DEFINED",
    "DEFINITION",
    "DELETE",
    "DELIMITED",
    "DEPTH",
    "DEREF",
    "DESC",
    "DESCRIBE",
    "DESCRIPTOR",
    "DETACH",
    "DETERMINISTIC",
    "DIAGNOSTICS",
    "DIRECTORIES",
    "DISABLE",
    "DISCONNECT",
    "DISTINCT",
    "DISTRIBUTE",
    "DO",
    "DOMAIN",
    "DOUBLE",
    "DROP",
    "DUMP",
    "DURATION",
    "DYNAMIC",
    "EACH",
    "ELEMENT",
    "ELSE",
    "ELSEIF",
    "EMPTY",
    "ENABLE",
    "END",
    "EQUAL",
    "EQUALS",
    "ERROR",
    "ESCAPE",
    "ESCAPED",
    "EVAL",
    "EVALUATE",
    "EXCEEDED",
    "EXCEPT",
    "EXCEPTION",
    "EXCEPTIONS",
    "EXCLUSIVE",
    "EXEC",
    "EXECUTE",
    "EXISTS",
    "EXIT",
    "EXPLAIN",
    "EXPLODE",
    "EXPORT",
    "EXPRESSION",
    "EXTENDED",
    "EXTERNAL",
    "EXTRACT",
    "FAIL",
    "FALSE",
    "FAMILY",
    "FETCH",
    "FIELDS",
    "FILE",
    "FILTER",
    "FILTERING",
    "FINAL",
    "FINISH",
    "FIRST",
    "FIXED",
    "FLATTERN",
    "FLOAT",
    "FOR",
    "FORCE",
    "FOREIGN",
    "FORMAT",
    "FORWARD",
    "FOUND",
    "FREE",
    "FROM",
    "FULL",
    "FUNCTION",
    "FUNCTIONS",
    "GENERAL",
    "GENERATE",
    "GET",
    "GLOB",
    "GLOBAL",
    "GO",
    "GOTO",
    "GRANT",
    "GREATER",
    "GROUP",
    "GROUPING",
    "HANDLER",
    "HASH",
    "HAVE",
    "HAVING",
    "HEAP",
    "HIDDEN",
    "HOLD",
    "HOUR",
    "IDENTIFIED",
    "IDENTITY",
    "IF",
    "IGNORE",
    "IMMEDIATE",
    "IMPORT",
    "IN",
    "INCLUDING",
    "INCLUSIVE",
    "INCREMENT",
    "INCREMENTAL",
    "INDEX",
    "INDEXED",
    "INDEXES",
    "INDICATOR",
    "INFINITE",
    "INITIALLY",
    "INLINE",
    "INNER",
    "INNTER",
    "INOUT",
    "INPUT",
    "INSENSITIVE",
    "INSERT",
    "INSTEAD",
    "INT",
    "INTEGER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "INVALIDATE",
    "IS",
    "ISOLATION",
    "ITEM",
    "ITEMS",
    "ITERATE",
    "JOIN",
    "KEY",
    "KEYS",
    "LAG",
    "LANGUAGE",
    "LARGE",
    "LAST",
    "LATERAL",
    "LEAD",
    "LEADING",
    "LEAVE",
    "LEFT",
    "LENGTH",
    "LESS",
    "LEVEL",
    "LIKE",
    "LIMIT",
    "LIMITED",
    "LINES",
    "LIST",
    "LOAD",
    "LOCAL",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "LOCATION",
    "LOCATOR",
    "LOCK",
    "LOCKS",
    "LOG",
    "LOGED",
    "LONG",
    "LOOP",
    "LOWER",
    "MAP",
    "MATCH",
    "MATERIALIZED",
    "MAX",
    "MAXLEN",
    "MEMBER",
    "MERGE",
    "METHOD",
    "METRICS",
    "MIN",
    "MINUS",
    "MINUTE",
    "MISSING",
    "MOD",
    "MODE",
    "MODIFIES",
    "MODIFY",
    "MODULE",
    "MONTH",
    "MULTI",
    "MULTISET",
    "NAME",
    "NAMES",
    "NATIONAL",
    "NATURAL",
    "NCHAR",
    "NCLOB",
    "NEW",
    "NEXT",
    "NO",
    "NONE",
    "NOT",
    "NULL",
    "NULLIF",
    "NUMBER",
    "NUMERIC",
    "OBJECT",
    "OF",
    "OFFLINE",
    "OFFSET",
    "OLD",
    "ON",
    "ONLINE",
    "ONLY",
    "OPAQUE",
    "OPEN",
    "OPERATOR",
    "OPTION",
    "OR",
    "ORDER",
    "ORDINALITY",
    "OTHER",
    "OTHERS",
    "OUT",
    "OUTER",
    "OUTPUT",
    "OVER",
    "OVERLAPS",
    "OVERRIDE",
    "OWNER",
    "PAD",
    "PARALLEL",
    "PARAMETER",
    "PARAMETERS",
    "PARTIAL",
    "PARTITION",
    "PARTITIONED",
    "PARTITIONS",
    "PATH",
    "PERCENT",
    "PERCENTILE",
    "PERMISSION",
    "PERMISSIONS",
    "PIPE",
    "PIPELINED",
    "PLAN",
    "POOL",
    "POSITION",
    "PRECISION",
    "PREPARE",
    "PRESERVE",
    "PRIMARY",
    "PRIOR",
    "PRIVATE",
    "PRIVILEGES",
    "PROCEDURE",
    "PROCESSED",
    "PROJECT",
    "PROJECTION",
    "PROPERTY",
    "PROVISIONING",
    "PUBLIC",
    "PUT",
    "QUERY",
    "QUIT",
    "QUORUM",
    "RAISE",
    "RANDOM",
    "RANGE",
    "RANK",
    "RAW",
    "READ",
    "READS",
    "REAL",
    "REBUILD",
    "RECORD",
    "RECURSIVE",
    "REDUCE",
    "REF",
    "REFERENCE",
    "REFERENCES",
    "REFERENCING",
    "REGEXP",
    "REGION",
    "REINDEX",
    "RELATIVE",
    "RELEASE",
    "REMAINDER",
    "RENAME",
    "REPEAT",
    "REPLACE",
    "REQUEST",
    "RESET",
    "RESIGNAL",
    "RESOURCE",
    "RESPONSE",
    "RESTORE",
    "RESTRICT",
    "RESULT",
    "RETURN",
    "RETURNING",
    "RETURNS",
    "REVERSE",
    "REVOKE",
    "RIGHT",
    "ROLE",
    "ROLES",
    "ROLLBACK",
    "ROLLUP",
    "ROUTINE",
    "ROW",
    "ROWS",
    "RULE",
    "RULES",
    "SAMPLE",
    "SATISFIES",
    "SAVE",
    "SAVEPOINT",
    "SCAN",
    "SCHEMA",
    "SCOPE",
    "SCROLL",
    "SEARCH",
    "SECOND",
    "SECTION",
    "SEGMENT",
    "SEGMENTS",
    "SELECT",
    "SELF",
    "SEMI",
    "SENSITIVE",
    "SEPARATE",
    "SEQUENCE",
    "SERIALIZABLE",
    "SESSION",
    "SET",
    "SETS",
    "SHARD",
    "SHARE",
    "SHARED",
    "SHORT",
    "SHOW",
    "SIGNAL",
    "SIMILAR",
    "SIZE",
    "SKEWED",
    "SMALLINT",
    "SNAPSHOT",
    "SOME",
    "SOURCE",
    "SPACE",
    "SPACES",
    "SPARSE",
    "SPECIFIC",
    "SPECIFICTYPE",
    "SPLIT",
    "SQL",
    "SQLCODE",
    "SQLERROR",
    "SQLEXCEPTION",
    "SQLSTATE",
    "SQLWARNING",
    "START",
    "STATE",
    "STATIC",
    "STATUS",
    "STORAGE",
    "STORE",
    "STORED",
    "STREAM",
    "STRING",
    "STRUCT",
    "STYLE",
    "SUB",
    "SUBMULTISET",
    "SUBPARTITION",
    "SUBSTRING",
    "SUBTYPE",
    "SUM",
    "SUPER",
    "SYMMETRIC",
    "SYNONYM",
    "SYSTEM",
    "TABLE",
    "TABLESAMPLE",
    "TEMP",
    "TEMPORARY",
    "TERMINATED",
    "TEXT",
    "THAN",
    "THEN",
    "THROUGHPUT",
    "TIME",
    "TIMESTAMP",
    "TIMEZONE",
    "TINYINT",
    "TO",
    "TOKEN",
    "TOTAL",
    "TOUCH",
    "TRAILING",
    "TRANSACTION",
    "TRANSFORM",
    "TRANSLATE",
    "TRANSLATION",
    "TREAT",
    "TRIGGER",
    "TRIM",
    "TRUE",
    "TRUNCATE",
    "TTL",
    "TUPLE",
    "TYPE",
    "UNDER",
    "UNDO",
    "UNION",
    "UNIQUE",
    "UNIT",
    "UNKNOWN",
    "UNLOGGED",
    "UNNEST",
    "UNPROCESSED",
    "UNSIGNED",
    "UNTIL",
    "UPDATE",
    "UPPER",
    "URL",
    "USAGE",
    "USE",
    "USER",
    "USERS",
    "USING",
    "UUID",
    "VACUUM",
    "VALUE",
    "VALUED",
    "VALUES",
    "VARCHAR",
    "VARIABLE",
    "VARIANCE",
    "VARINT",
    "VARYING",
    "VIEW",
    "VIEWS",
    "VIRTUAL",
    "VOID",
    "WAIT",
    "WHEN",
    "WHENEVER",
    "WHERE",
    "WHILE",
    "WINDOW",
    "WITH",
    "WITHIN",
    "WITHOUT",
    "WORK",
    "WRAPPED",
    "WRITE",
    "YEAR",
    "ZONE",
]



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, FrozenSet

from .error import *  # noqa

if TYPE_CHECKING:
    from .connection import Connection

__version__: str = "0.7.4"

# Globals https://www.python.org/dev/peps/pep-0249/#globals
apilevel: str = "2.0"
threadsafety: int = 3
paramstyle: str = "qmark"


class DBAPITypeObject(FrozenSet[str]):
    """Type Objects and Constructors

    https://www.python.org/dev/peps/pep-0249/#type-objects-and-constructors
    """

    def __eq__(self, other: object):
        if isinstance(other, frozenset):
            return frozenset.__eq__(self, other)
        else:
            return other in self

    def __ne__(self, other: object):
        if isinstance(other, frozenset):
            return frozenset.__ne__(self, other)
        else:
            return other not in self

    def __hash__(self):
        return frozenset.__hash__(self)


# https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/HowItWorks.NamingRulesDataTypes.html
STRING: DBAPITypeObject = DBAPITypeObject(("STRING",))
BINARY: DBAPITypeObject = DBAPITypeObject(("STRING",))
BOOLEAN: DBAPITypeObject = DBAPITypeObject(
    (
        "BOOL",
        "NULL",
    )
)
NUMBER: DBAPITypeObject = DBAPITypeObject(("NUMBER",))
JSON: DBAPITypeObject = DBAPITypeObject(("STRING",))


def connect(*args, **kwargs) -> "Connection":
    from .connection import Connection
    from .superset_dynamodb.pydynamodb import SupersetCursor

    connector = kwargs.get("connector", None)
    if connector is not None and connector.lower() == "superset":
        kwargs.update({"cursor_class": SupersetCursor})

    return Connection(*args, **kwargs)



# -*- coding: utf-8 -*-
import os
import logging
import importlib
from abc import ABCMeta
from typing import Optional, Union

from .querydb import (
    QueryDBConfig,
    QueryDB,
    SUPPORTED_QUERYDB_CONFIG,
    DEFAULT_QUERYDB_TYPE,
    DEFAULT_QUERYDB_URL,
    DEFAULT_QUERYDB_LOAD_BATCH_SIZE,
    DEFAULT_QUERYDB_EXPIRE_TIME,
    DEFAULT_QUERYDB_PURGE_TIME,
)
from ..model import Statement
from ..util import strtobool
from ..error import NotSupportedError, OperationalError
from .querydb_sqlite import SqliteMemQueryDB, SqliteFileQueryDB

_logger = logging.getLogger(__name__)  # type: ignore


class QueryDBHelper(metaclass=ABCMeta):
    @staticmethod
    def create(statement: Statement, config: QueryDBConfig = None, **kwargs) -> QueryDB:
        if config is None:
            _config = QueryDBHelper.get_default_config(**kwargs)
        else:
            _config = config

        _query_db_class = None
        if _config.db_type.lower() == DEFAULT_QUERYDB_TYPE:
            _query_db_class = (
                SqliteMemQueryDB
                if _config.db_url.lower() == DEFAULT_QUERYDB_URL
                else SqliteFileQueryDB
            )
        else:
            if _config.db_class:
                db_apis = _config.db_class.split(":")
                if len(db_apis) == 2:
                    _query_db_module = importlib.import_module(db_apis[0])
                    _query_db_class = getattr(_query_db_module, db_apis[1])
                else:
                    raise OperationalError("QueryDB class is invalid.")
            else:
                raise NotSupportedError("QueryDB class is not specified.")

        _query_db = _query_db_class(statement, _config, **kwargs)
        if _query_db is None:
            raise OperationalError("QueryDB is not specified.")
        else:
            return _query_db

    @staticmethod
    def get_default_config(**kwargs):
        db_type = QueryDBHelper.get_config_value("querydb_type", **kwargs)
        db_url = QueryDBHelper.get_config_value("querydb_url", **kwargs)
        db_class = QueryDBHelper.get_config_value("querydb_class", **kwargs)
        load_batch_size = QueryDBHelper.get_config_value(
            "querydb_load_batch_size", **kwargs
        )
        expire_time = QueryDBHelper.get_config_value("querydb_expire_time", **kwargs)
        purge_enabled = QueryDBHelper.get_config_value(
            "querydb_purge_enabled", **kwargs
        )
        purge_time = QueryDBHelper.get_config_value("querydb_purge_time", **kwargs)

        _db_type = db_type if db_type is not None else DEFAULT_QUERYDB_TYPE
        _db_url = db_url if db_url is not None else DEFAULT_QUERYDB_URL
        _db_class = db_class
        _load_batch_size = (
            int(load_batch_size)
            if load_batch_size is not None
            else DEFAULT_QUERYDB_LOAD_BATCH_SIZE
        )
        _expire_time = (
            int(expire_time) if expire_time is not None else DEFAULT_QUERYDB_EXPIRE_TIME
        )
        _purge_enabled = strtobool(purge_enabled) if purge_enabled is not None else True
        _purge_time = (
            int(purge_time) if purge_time is not None else DEFAULT_QUERYDB_PURGE_TIME
        )

        return QueryDBConfig(
            _db_type,
            _db_class,
            _db_url,
            load_batch_size=_load_batch_size,
            expire_time=_expire_time,
            purge_enabled=_purge_enabled,
            purge_time=_purge_time,
        )

    @staticmethod
    def get_config_value(config_name: str, **kwargs) -> Optional[Union[str, int]]:
        if config_name in kwargs:
            return kwargs[config_name]
        else:
            if config_name in SUPPORTED_QUERYDB_CONFIG:
                (env_name, default_val) = SUPPORTED_QUERYDB_CONFIG[config_name]
                env_val = os.getenv(env_name, None)
                if env_val is None:
                    return default_val
                else:
                    return env_val
            else:
                return None



# -*- coding: utf-8 -*-
import logging
import sqlean as sqlite3
from datetime import datetime, date
from contextlib import closing
from typing import Any, Type

from .querydb import QueryDB, QueryDBConfig
from ..model import Statement

_logger = logging.getLogger(__name__)  # type: ignore


class SqliteFileQueryDB(QueryDB):
    def __init__(
        self,
        statement: Statement,
        config: QueryDBConfig,
        **kwargs,
    ) -> None:
        super().__init__(statement, config, **kwargs)
        self._connection = None

    @property
    def connection(self):
        if self._connection is None:
            sqlite3.extensions.enable("math", "regexp", "stats", "text")
            self._connection = sqlite3.connect(
                self.config.db_url,
                detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
            )

        return self._connection

    def type_conversion(self, type: Type[Any]) -> str:
        type_mapping = {
            str: "TEXT",
            int: "INTEGER",
            float: "REAL",
            bytes: "TEXT",
            bool: "INTEGER",
            datetime: "TIMESTAMP",
            date: "DATE",
        }

        return type_mapping.get(type, "TEXT")

    def has_table(self, table: str) -> bool:
        with closing(self.connection.cursor()) as cursor:
            cursor.execute(
                "SELECT count(name) FROM sqlite_master WHERE type='table' AND name=?",
                (table,),
            )
            result = cursor.fetchone()
            if result and result == (1,):
                return True

        return False

    def purge(self) -> int:
        purged_count = super().purge()

        if purged_count > 0:
            self.connection.execute("VACUUM")

        return purged_count


class SqliteMemQueryDB(SqliteFileQueryDB):
    def __init__(
        self,
        statement: Statement,
        config: QueryDBConfig,
        **kwargs,
    ) -> None:
        kwargs.update({"cache_enabled": False})
        super().__init__(statement, config, **kwargs)
